#ifndef __fnccommon_h_
#define __fnccommon_h_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "../include/fnccheck.h"



/** used MACROS and other DEFINES **/
/* the MAX/MIN macro */
#define HX_MAX(a,b)  (((a)>(b))?(a):(b))
#define HX_MIN(a,b)  (((a)<(b))?(a):(b))
#define EPS 0.000001  /* 1 usec */
#define ABS(v) ((v)<0?-(v):(v))
/* macros to display (potentially null) names */
/* function name */
#define HX_DNAME(n)  (((n)==NULL)?"??":(n))
/* file */
#define HX_DFILE(n)  (((n)==NULL)?"??":(n))
/* object */
#define HX_DOBJ(n)   (((n)==NULL)?"??.so":(n))
/* string buffers */
#define HX_MAX_BUFFER 1024
/* the kind of time used  -> hum... this is duplicate in lib/... */
#define HX_TIME_EXT 0   /* absolute time */
#define HX_TIME_CPU 1   /* internal time (CPU) */
#define HX_TIME_SYS 2   /* CPU+SYSTEM times */
/* temp file for 'nm' command output (is there a better/nicer way ?) */
#define HX_NM_FTEMP  "%s/.fnccheck_nm_temp_%d.txt"
#define HX_NM_FTEMP2 "%s/.fnccheck_nm2_temp_%d.txt"

/** structures needed by the program **/

/* Description of a function or call address */
/* This structure is overlaid on some of the others. */
typedef struct
{
   void *addr;		/* address of event */
   char *name;		/* name of function containing address */
   char *object;	/* path to object containing address */
   char *file;		/* source file containing the function */
   int line;		/* line number in source file */
   int valid;		/* is this name ok? */
}HX_Loc;

/* table of stats for functions */
/*    structure of a function   */
/* not really nice. The max number of callers for each
   function that are registered. I use this to prevent
   reallocs during profile (to have a constant overcost). */
#define HX_MAX_CALLER 64
typedef struct
{
  void *function;  /* the pointer to the function */
  char *name;
  char *object;
  char *file;      /* the FILE:LINE info for the function */
  int line;
  int valid;    /* do the name is okay ? */
  unsigned long int nb_calls; /* total # of calls */
  unsigned long int nb_sec;   /* total # of secs */
  unsigned long int nb_usec;  /* total # of usecs */
  /* local time spend in function */
  unsigned long int nb_locsec;   /* # of secs */
  unsigned long int nb_locusec;  /* # of usecs */
  /* min and max time for the function */
  unsigned long int min_sec;   /* # of secs */
  unsigned long int min_usec;       /* # of usecs */
  unsigned long int max_sec;   /* # of secs */
  unsigned long int max_usec;       /* # of usecs */
  /* computed values */
  double local;
  double total;
  double min, max;
  void *called_by[HX_MAX_CALLER];
  HX_Loc rcb[HX_MAX_CALLER];	    /* datas about real calls */
  unsigned int nb_called_by[HX_MAX_CALLER];
  unsigned int nb_of_callers;
  unsigned int my_index;
}HX_FDInfo;

/* this is for memory files */
/* structure for stack of calls */
#define HX_MemStack	HX_Loc

typedef struct HX_Mem
{
  void *ptr;
  unsigned int size_a;
  unsigned int block;

  void *where_a;   /* allocation place */
  char *func_a;    /* corresp. function */
  char *object_a;    /* corresp. object */
  char *file_a;    /* corresp. file:line */
  int line_a;
  int valid_a;

  /* call stack */
  HX_MemStack *stack_a;
  unsigned int nb_stack_a;

  void *where_f;   /* freeing place */
  char *func_f;    /* corresp. function */
  char *object_f;  /* corresp. object */
  char *file_f;    /* corresp. file:line */
  int line_f;
  int valid_f;

  /* stack at the free */
  HX_MemStack *stack_f;
  unsigned int nb_stack_f;

  struct HX_Mem *rnext;	/* realloc list */
  struct HX_Mem *rprev;	/* prior alloc */
  struct HX_Mem *anext;	/* next at this addr */
}HX_Mem;

/* structure of a node (for call-graph) */
typedef struct _HX_Node HX_Node;
struct _HX_Node
{
  HX_FDInfo *function;  /* corresponding function */
  HX_Node **nexts;      /* list of childs (ended by NULL) */
  HX_Node **prevs;      /* list of callers (ended by NULL) */
  int keep;             /* bool: if true, the function is keeped */
  int treated;          /* bool+: used to propagate "keep" */
  int tag;              /* tag for cycle detection */
};

/* a dynamic object */
typedef struct
{
  void *base;
  char path[1];
}HX_DynObj;





/** variables **/
extern int hx_time_type;
extern HX_FDInfo *hx_dfinfo;
extern int hx_nb_dfinfo;
extern int hx_nb_objects;
extern HX_DynObj **hx_objects;





#endif /*__fnccommon_h_*/
