#include <signal.h>
#include "fnccommon.h"
#include "fnctools.h"
#include "fncgraph.h"
#include "fncmem.h"
#include "fncnames.h"
#include "fncfunctions.h"


/* temporary (a 'core dump' tracking :o) */
#define PRINTF(m) printf(m); fflush(stdout)


/* the output file */
static char *HX_DROP_FILE=FNCCHK_DEFAULT_FILE;


int hx_time_type=HX_TIME_EXT;


/* from fncmem.c */
/* memory count. after counting, this is the memory leak */
extern long hx_nb_mem_used;
extern long hx_nb_mem_max;  /* gives the real MAX memory used */
/* length of the memory used */
extern long hx_nb_heap_size;


HX_DynObj **hx_objects;
int hx_nb_objects;

/* mult for sort order */
int hx_mult_for_sort=-1;


/* a little string set by functions to indicate where we are */
char *hx_where_am_i="initializations";
#define hx_set_message(msg) hx_where_am_i = msg
/* this function is called if someone kills us */
/* I added this because memory treatments can be (very) long,
   and in this case it indicates to the user that we where not
   in a 'core dumping' phase, but in a normal treatment, */
void hx_damned(int dummy)
{
  hxmessage("killing signal received.");
  hxmessage("treatments where in:");
  hxmessage("     %s", hx_where_am_i);
  hxmessage("exiting.\n");
  exit(1);
}





/* sort functions */
int hx_sort_by_name(const void*el1, const void*el2)
{
  HX_FDInfo *rel1, *rel2;

  rel1 = (HX_FDInfo*)el1;
  rel2 = (HX_FDInfo*)el2;
  
  return(hx_mult_for_sort*strcmp(rel1->name, rel2->name));
}
int hx_sort_by_calls(const void*el1, const void*el2)
{
  HX_FDInfo *rel1, *rel2;

  rel1 = (HX_FDInfo*)el1;
  rel2 = (HX_FDInfo*)el2;

  if (rel1->nb_calls < rel2->nb_calls)
    return(hx_mult_for_sort*(-1));
  if (rel1->nb_calls > rel2->nb_calls)
    return(hx_mult_for_sort*1);
  return(0);
}
int hx_sort_by_local(const void*el1, const void*el2)
{
  HX_FDInfo *rel1, *rel2;

  rel1 = (HX_FDInfo*)el1;
  rel2 = (HX_FDInfo*)el2;

  if (rel1->local < rel2->local)
    return(hx_mult_for_sort*(-1));
  if (rel1->local > rel2->local)
    return(hx_mult_for_sort*1);
  return(0);  /* may never occur on reals */
}
int hx_sort_by_total(const void*el1, const void*el2)
{
  HX_FDInfo *rel1, *rel2;

  rel1 = (HX_FDInfo*)el1;
  rel2 = (HX_FDInfo*)el2;

  if (rel1->total < rel2->total)
    return(hx_mult_for_sort*(-1));
  if (rel1->total > rel2->total)
    return(hx_mult_for_sort*1);
  return(0);  /* may never occur on reals */
}




/* these functions cumulate the profile data from two .out files
   into a single, summed profile file. Please be sure that the
   two files come from the SAME EXECUTABLE (same program, same
   compilation). Else, unpredictable results in output will
   occur (most of time: many unresolved function names and
   bad cumulation of times                                   */
typedef struct
{
  void *func;
  unsigned long int sec, usec;
  unsigned long int lsec, lusec;
  unsigned long int misec, miusec;
  unsigned long int masec, mausec;
  unsigned long int nbc;
  void *callers[HX_MAX_CALLER];
  void *rcallers[HX_MAX_CALLER];
  unsigned int nb_called_by[HX_MAX_CALLER];
  unsigned int nb_of_callers;
}HX_CumFunction;
void hx_register_caller(HX_CumFunction *who, void *from, void *rfrom, int nb)
{
  unsigned int i;

  if (who->nb_of_callers >= HX_MAX_CALLER)
    return; /* list is full */
  for(i=0; i<who->nb_of_callers; i++)
    {
    if (who->callers[i] == from)
      return;  /* still referenced */
    }
  who->callers[who->nb_of_callers] = from;
  who->rcallers[who->nb_of_callers] = rfrom;
  who->nb_called_by[who->nb_of_callers] = nb;
  who->nb_of_callers++;
}
int hx_cumulate_profile(char *p1, char *p2, char *op)
{
  FILE *f1, *f2, *of;
  char buf1[HX_MAX_BUFFER], buf2[HX_MAX_BUFFER];
  char name1[HX_MAX_BUFFER], name2[HX_MAX_BUFFER];
  int i, j, k, curf, tm1, tm2, add;
  unsigned int n1, n2;
  HX_CumFunction *cumf1, *cumf2, *cumr;  /* list of functions */
  void *fnc;
  unsigned long int itemp2;
  unsigned int nr1, nr2, ssu1, ssu2, nf1, nf2;

  /* open the files */
  if ((f1 = fopen(p1, "r")) == NULL)
    {
    fprintf(stderr, "fncdump: cant open file '%s'.\n", p1);
    return(0);
    }
  if ((f2 = fopen(p2, "r")) == NULL)
    {
    fclose(f1);
    fprintf(stderr, "fncdump: cant open file '%s'.\n", p2);
    return(0);
    }
  /* check headers */
  fscanf(f1, "%1023s", buf1);
  fscanf(f2, "%1023s", buf2);
  if (strcmp(buf1, FNCCHK_HEADER) != 0)
    {
    fprintf(stderr, "fncdump: invalid header file in '%s'.\n", p1);
    fclose(f1);
    fclose(f2);
    return(0);
    }
  if (strcmp(buf2, FNCCHK_HEADER) != 0)
    {
    fprintf(stderr, "fncdump: invalid header file in '%s'.\n", p2);
    fclose(f1);
    fclose(f2);
    return(0);
    }
  /* number of elements */
  fscanf(f1, "%u", &n1);
  fscanf(f2, "%u", &n2);
  if ((n1 <= 0)||(n2 <= 0))
    {
    fprintf(stderr, "fncdump: invalid number of elements (%u||%u).\n", n1, n2);
    fclose(f1);
    fclose(f2);
    return(0);
    }
  /* alloc tables */
  if ((cumf1 = malloc(sizeof(HX_CumFunction)*n1)) == NULL)
    {
    fprintf(stderr, "fncdump: Memory error (%d bytes)!\n", sizeof(HX_CumFunction)*n1);
    fclose(f1);
    fclose(f2);
    return(0);
    }
  if ((cumf2 = malloc(sizeof(HX_CumFunction)*n2)) == NULL)
    {
    fprintf(stderr, "fncdump: Memory error (%d bytes)!\n", sizeof(HX_CumFunction)*n2);
    free(cumf1);
    fclose(f1);
    fclose(f2);
    return(0);
    }
  /* read elements */
  for(i=0; i<n1; i++)
    {
    fscanf(f1, "%p", &fnc);
    cumf1[i].func   = fnc;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].nbc    = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].sec    = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].usec   = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].lsec   = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].lusec  = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].misec  = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].miusec = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].masec  = (unsigned long int)itemp2;
    fscanf(f1, "%lu", &itemp2);
    cumf1[i].mausec = (unsigned long int)itemp2;
    }
  fscanf(f1, "%u", &nr1);
  fscanf(f1, "%u", &ssu1);
  fscanf(f1, "%u", &nf1);
  for(i=0; i<n1; i++)
    {
    fscanf(f1, "%u", &(cumf1[i].nb_of_callers));
    for(j=0; j<cumf1[i].nb_of_callers; j++)
      {
      fscanf(f1, "%p", &(cumf1[i].callers[j]));
      fscanf(f1, "%p", &(cumf1[i].rcallers[j]));
      fscanf(f1, "%x", &(cumf1[i].nb_called_by[j]));
      }
    }
  for(i=0; i<n2; i++)
    {
    fscanf(f2, "%p", &fnc);
    cumf2[i].func   = fnc;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].nbc    = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].sec    = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].usec   = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].lsec   = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].lusec  = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].misec  = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].miusec = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].masec  = (unsigned long int)itemp2;
    fscanf(f2, "%lu", &itemp2);
    cumf2[i].mausec = (unsigned long int)itemp2;
    }
  fscanf(f2, "%u", &nr2);
  fscanf(f2, "%u", &ssu2);
  fscanf(f2, "%u", &nf2);
  for(i=0; i<n2; i++)
    {
    fscanf(f2, "%u", &(cumf2[i].nb_of_callers));
    for(j=0; j<cumf2[i].nb_of_callers; j++)
      {
      fscanf(f2, "%p", &(cumf2[i].callers[j]));
      fscanf(f2, "%p", &(cumf2[i].rcallers[j]));
      fscanf(f2, "%x", &(cumf2[i].nb_called_by[j]));
      }
    }
  /* read the time modes */
  fscanf(f1, "%d", &tm1);
  fscanf(f2, "%d", &tm2);
  /* must be the same */
  if (tm1 != tm2)
    {
    fprintf(stderr,
      "fncdump: time mode in files %s and %s are not the same.\n", p1, p2);
    fclose(f1);
    fclose(f2);
    free(cumf1);
    free(cumf2);
    return(0);
    }
  /* read exec names (not implemented) */
  fscanf(f1, "%1023s", name1);
  fscanf(f2, "%1023s", name2);
  /* files readen */
  fclose(f1);
  fclose(f2);
  /* must be the same exec */
  if (strcmp(name1, name2) != 0)
    {
    fprintf(stderr,
      "fncdump: exec name in files %s and %s are not the same.\n", p1, p2);
    free(cumf1);
    free(cumf2);
    return(0);
    }
  
  /* now we duplicate the functions from #1 into a n1+n2 list
     (max of possible functions)  */
  if ((cumr = malloc(sizeof(HX_CumFunction)*(n1+n2))) == NULL)
    {
    fprintf(stderr, "fncdump: Memory error (%d bytes)!\n", sizeof(HX_CumFunction)*(n1+n2));
    free(cumf1);
    free(cumf2);
    return(0);
    }
  for(i=0; i<n1; i++)
    {
    cumr[i] = cumf1[i];
    }
  /* for each function from #2:  if not registered, duplicate
     it in the new list.  if registered, add its stats to the
     existing list */
  curf = n1;
  for(i=0; i<n2; i++)
    {
    /* search the function */
    add = 0;
    for(j=0; j<n1; j++)
      {
      if (cumr[j].func == cumf2[i].func)
        {
        add = 1;
        break;
        }
      }
    if (add)
      {/* add stats */
      /* # of call */
      cumr[j].nbc += cumf2[i].nbc;
      /* times */
      cumr[j].sec    += cumf2[i].sec;
      cumr[j].usec   += cumf2[i].usec;
      cumr[j].lsec   += cumf2[i].lsec;
      cumr[j].lusec  += cumf2[i].lusec;
      cumr[j].misec  += cumf2[i].misec;
      cumr[j].miusec += cumf2[i].miusec;
      cumr[j].masec  += cumf2[i].masec;
      cumr[j].mausec += cumf2[i].mausec;
      /* callers list */
      for(k=0; k<cumf2[i].nb_of_callers; k++)
        {
        hx_register_caller(&(cumr[j]), cumf2[i].callers[k],
	  cumf2[i].rcallers[k], cumf2[i].nb_called_by[k]);
        }
      }
    else
      {/* duplicate element */
      cumr[curf] = cumf2[i];
      curf++;
      }
    }

  /* now we just output the new file */
  if (strcmp(op, "-") == 0)
    {
    of = stdout;
    }
  else
    {
    if ((of = fopen(op, "w")) == NULL)
      {
      fprintf(stderr, "fncdump: cant create output file '%s'.\n", op);
      free(cumf1);
      free(cumf2);
      free(cumr);
      }
    }
  fprintf(of, "%s\n", FNCCHK_HEADER);
  fprintf(of, "%d\n", curf);
  for(i=0; i<curf; i++)
    {
    fprintf(of, "%p %lu %lu %lu %lu %lu %lu %lu %lu %lu\n",
                  cumr[i].func, cumr[i].nbc,
                  cumr[i].sec, cumr[i].usec,
                  cumr[i].lsec, cumr[i].lusec,
                  cumr[i].misec, cumr[i].miusec,
                  cumr[i].masec, cumr[i].mausec);
    }
  fprintf(of, "%d\n", nr1+nr2); /* sum of reallocs */
  fprintf(of, "%u %u\n", HX_MAX(ssu1, ssu2), curf);
  for(i=0; i<curf; i++)
    {
    fprintf(of, "%u ", cumr[i].nb_of_callers);
    for(j=0; j<cumr[i].nb_of_callers; j++)
      {
      fprintf(of, "%p %p %#x", cumr[i].callers[j], cumr[i].rcallers[j],
        cumr[i].nb_called_by[j]);
      if (j+1 < cumr[i].nb_of_callers)
        putc(' ', of);
      }
    putc('\n', of);
    }
  fprintf(of, "%d\n", tm1);  /* time mode */
  fprintf(of, "%s\n", name1); /* exec name */

  fclose(of);  

  /* leaving */
  free(cumf1);
  free(cumf2);
  free(cumr);
  return(1);  /* ok */
}


#define IFDECO  if (options & FNCCHK_NO_DECORATION)
#define IFNDECO if (!(options & FNCCHK_NO_DECORATION))
/* display the flat profile */
int hx_display_profile(int options, unsigned int nbinfo, double total)
{
  int i;
  double cum=0.;

  if (options & FNCCHK_NO_PROFILE)
    return(1);

  if (!(options & FNCCHK_NO_DECORATION))
    {
    printf(HX_TTL_FLAT);
    printf("\n");
    if (options & FNCCHK_CUMUL)
      {
      printf("|%17s|%5s|%17s|%10s|%11s|\n", "local      ", "cum. ", "total      ", "", "");
      printf("|%11s|%5s|%5s|%11s|%5s|%10s|%11s| name\n", "sec. ", "%  ", "%  ", "sec. ", "%  ", "calls    ", "t. sec/call");
      printf("|-----------|-----|-----|-----------|-----|----------|-----------|--------\n");
      }
    else
      {
      printf("|%17s|%17s|%10s|%11s|\n", "local      ", "total      ", "", "");
      printf("|%11s|%5s|%11s|%5s|%10s|%11s| name\n", "sec. ", "%  ", "sec. ", "%  ", "calls    ", "t. sec/call");
      printf("|-----------|-----|-----------|-----|----------|-----------|--------\n");
      }
    }
  else
    {
    printf(">profile\n");
    }
  for(i=0; i<nbinfo; i++)
    {
    if (!(((options & FNCCHK_NO_UNDEF) != 0)&&(!hx_dfinfo[i].valid)))
      {
      if (!(((options & FNCCHK_NO_SPONTANEOUS) != 0)&&
           (hx_dfinfo[i].nb_of_callers == 0)&&
	   (strcmp(hx_dfinfo[i].name, "main") != 0)))
        {
	cum += hx_dfinfo[i].local;
	if (options & FNCCHK_NO_DECORATION)
	  {
	  if (!(options & FNCCHK_CUMUL))
	    printf("%f %f %f %f %u %f \"%s\"\n",
	       hx_dfinfo[i].local, total==0.?0.:100.*hx_dfinfo[i].local/total,
	       hx_dfinfo[i].total, total==0.?0.:100.*hx_dfinfo[i].total/total,
	       (unsigned int)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].total / (double)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].name);
	  else
	    printf("%f %f %f %f %f %u %f \"%s\"\n",
	       hx_dfinfo[i].local, total==0.?0.:100.*hx_dfinfo[i].local/total,
	       total==0.?0.:100.*cum/total,
	       hx_dfinfo[i].total, total==0.?0.:100.*hx_dfinfo[i].total/total,
	       (unsigned int)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].total / (double)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].name);
	  }
	else
	  {
	  if (!(options & FNCCHK_CUMUL))
	    printf("|%11f|%5.1f|%11f|%5.1f|%10u|%11f| %s\n",
	       hx_dfinfo[i].local, total==0.?0.:100.*hx_dfinfo[i].local/total,
	       hx_dfinfo[i].total, total==0.?0.:100.*hx_dfinfo[i].total/total,
	       (unsigned int)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].total / (double)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].name);
	  else
	    printf("|%11f|%5.1f|%5.1f|%11f|%5.1f|%10u|%11f| %s\n",
	       hx_dfinfo[i].local, total==0.?0.:100.*hx_dfinfo[i].local/total,
	       total==0.?0.:100.*cum/total,
	       hx_dfinfo[i].total, total==0.?0.:100.*hx_dfinfo[i].total/total,
	       (unsigned int)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].total / (double)hx_dfinfo[i].nb_calls,
	       hx_dfinfo[i].name);
	  }
	}
      }
    }
  if (options & FNCCHK_NO_DECORATION)
    {
    printf("<profile\n");
    }
  else
    {
    printf("\n");
    }
  return(1);
}

int hx_display_minmax(int options, unsigned int nbinfo)
{
  int i;

  if (options & FNCCHK_NO_MINMAX)
    return(1);

  if (options & FNCCHK_NO_DECORATION)
    {
    printf(">minmax\n");
    }
  else
    {
    printf(HX_TTL_MINMAX);
    printf("\n");
    printf("|%4s|%13s|%13s|n!me\n", "indx", "MIN     ", "MAX     ");
    printf("|----|-------------|-------------|-----\n");
    }
  for(i=0; i<nbinfo; i++)
    {
    if (!(((options & FNCCHK_NO_SPONTANEOUS) != 0)&&
         (hx_dfinfo[i].nb_of_callers == 0)&&
	 (strcmp(hx_dfinfo[i].name, "main") != 0)))
      {
      if (options & FNCCHK_NO_DECORATION)
	{
	printf("%d %f %f \"%s\"\n",
	       i, hx_dfinfo[i].min, hx_dfinfo[i].max,
	       hx_dfinfo[i].name);
	}
      else
	{
	printf("|%4d|%13f|%13f| %s\n",
	       i, hx_dfinfo[i].min, hx_dfinfo[i].max,
	       hx_dfinfo[i].name);
	}
      }
    }
  if (options & FNCCHK_NO_DECORATION)
    {
    printf("<minmax\n");
    }
  else
    {
    printf("\n");
    }
  return(1);
}

void hx_format1(int i, int cb, int cs, char *name, int nl)
{
  char numbuf[16];

  sprintf(numbuf, "[%d]", i);
  printf("         %6d/%-6d %6s     %s", cb, cs, numbuf, name);
  if (nl) putchar('\n');
}

void hx_format0(int i, int cb, char *name)
{
  char numbuf[16];

  sprintf(numbuf, "[%d]", i);
  printf(" %-6s  %6d        %6s %s\n", numbuf, cb, numbuf, name);
}

int hx_display_calls(int options, unsigned int nbinfo, int nb_hide, int nb_spont)
{
  int i, j, k, fc, prevd;


  if (options & FNCCHK_NO_CALL)
    return(1);

  /* display who calls who */
  if (options & FNCCHK_NO_DECORATION)
    {
    printf(">callgraph\n");
    }
  else
    {
    printf(HX_TTL_CALLG);
    printf("\n");
    }
  for(i=0; i<nbinfo; i++)
    {
    if (((options & FNCCHK_NO_UNDEF) != 0)&&(!hx_dfinfo[i].valid))
      {
      nb_hide++;
      }
    else
      {
      if (((options & FNCCHK_NO_SPONTANEOUS) != 0)&&
          (hx_dfinfo[i].nb_of_callers == 0)&&
	  (strcmp(hx_dfinfo[i].name, "main") != 0))
        {
	nb_spont++;
	}
      else
        {
	prevd = 0;
	if (i > 0)
	  {
	  IFNDECO
	    printf("--------------------------------\n");
	  }
	if ((options & FNCCHK_CALLS_CALLED)||(!(options & FNCCHK_CALLS)))
	  {
	  prevd = 1;
	  IFDECO
	    {
	    printf("\"%s\" %d ", hx_dfinfo[i].name, i);
	    if (options & FNCCHK_CALLS_MORE)
	      printf("\"%s:%d\"{%s} ", ((hx_dfinfo[i].file!=NULL)?hx_dfinfo[i].file:"<n/a>"),
	                           hx_dfinfo[i].line, hx_dfinfo[i].object);
	    }
	  if (hx_dfinfo[i].nb_of_callers > 0)
	    {
            IFDECO
	      printf("%d\n", hx_dfinfo[i].nb_of_callers);
	    }
	  else
	    {
            IFDECO
	      printf("%d", hx_dfinfo[i].nb_of_callers);
            else
	      printf("spontaneously called.\n");
	    }
	  for(j=0; j<hx_dfinfo[i].nb_of_callers; j++)
	    {
	    k = hx_get_func_by_ptr(hx_dfinfo[i].called_by[j], nbinfo)-1;
	    if (options & FNCCHK_CALLS_MORE)
	      {
	      IFDECO
	        printf(" %d \"%s:%d\"{%s}", k,
	                              ((hx_dfinfo[i].rcb[j].file!=NULL)?hx_dfinfo[i].rcb[j].file:"<n/a>"),
				        hx_dfinfo[i].rcb[j].line, hx_dfinfo[i].rcb[j].object);
	      else
	        {
		hx_format1(k, hx_dfinfo[i].nb_called_by[j],
		  hx_dfinfo[i].nb_calls, hx_dfinfo[k].name, 0);
	        printf(" (%s:%d){%s}\n",
	          ((hx_dfinfo[i].rcb[j].file!=NULL)?hx_dfinfo[i].rcb[j].file:"<n/a>"),
		  hx_dfinfo[i].rcb[j].line, hx_dfinfo[i].rcb[j].object);
		}
	      }
	    else
	      IFDECO
	        printf(" %d", k);
	      else
		hx_format1(k, hx_dfinfo[i].nb_called_by[j],
		  hx_dfinfo[i].nb_calls, hx_dfinfo[k].name, 1);
	    }
	  IFDECO
	    printf("\n");
	  }
	IFNDECO
	  hx_format0(i, hx_dfinfo[i].nb_calls, hx_dfinfo[i].name);
	if ((options & FNCCHK_CALLS_CALLED)||(options & FNCCHK_CALLS)) /* who this function calls */
	  {
	  IFDECO
	    printf("\"%s\" %d ", hx_dfinfo[i].name, i);
	  if (options & FNCCHK_CALLS_MORE)
	    {
	    IFDECO
	      printf("\"%s:%d\"{%s} ", ((hx_dfinfo[i].file!=NULL)?hx_dfinfo[i].file:"<n/a>"),
	                hx_dfinfo[i].line, hx_dfinfo[i].object);
	    }
	  fc = 1;
	  for(j=0; j<nbinfo; j++)
	    {
	    int l = hx_get_func_by_ptr(hx_dfinfo[j].function, nbinfo)-1;
	    for(k=0; k<hx_dfinfo[j].nb_of_callers; k++)
	      {
	      if (hx_dfinfo[j].called_by[k] == hx_dfinfo[i].function)
		{
		if (fc == 1)
		  {
		  fc = 0;
		  IFDECO
		    printf("\n");
		  }
		if (options & FNCCHK_CALLS_MORE)
		  {
		  IFDECO
		    printf(" %d \"%s:%d\"{%s}", l,
	                        	   ((hx_dfinfo[j].rcb[k].file!=NULL)?hx_dfinfo[j].rcb[k].file:"<n/a>"),
					   hx_dfinfo[j].rcb[k].line, hx_dfinfo[j].rcb[k].object);
		  else
	            {
		    hx_format1(l, hx_dfinfo[j].nb_called_by[k],
		      hx_dfinfo[j].nb_calls, hx_dfinfo[l].name, 0);
	            printf(" (%s:%d){%s},\n",
	              ((hx_dfinfo[j].rcb[k].file!=NULL)?hx_dfinfo[j].rcb[k].file:"<n/a>"),
		      hx_dfinfo[j].rcb[k].line, hx_dfinfo[j].rcb[k].object);
		    }
		  }
		else
		  {
		  IFDECO
		    printf(" %d", l);
		  else
		    hx_format1(l, hx_dfinfo[j].nb_called_by[k],
		      hx_dfinfo[j].nb_calls, hx_dfinfo[l].name, 1);
		  }
		}
	      }
	    }
	  
	  if (fc == 1)
	    {
	    IFDECO
	      {
	      if (options & FNCCHK_CALLS_MORE)
	        printf("\n * \"*\"");
	      else
	        printf("\n *");
	      }
	    }
	  IFDECO
	    printf("\n");
	  }
	}
      }
    }
  IFDECO
    {
    printf("<callgraph\n");
    }
  return(1);
}


int hx_display_functions(int options, unsigned int nbinfo)
{
  int i;

  if (!(options & FNCCHK_FUNC_MORE))
    return(1);

  IFDECO
    {
    printf(">details\n");
    }
  else
    {
    printf(HX_TTL_DETAIL);
    printf("\n");
    }
  for(i=0; i<nbinfo; i++)
    {
    if (!(((options & FNCCHK_NO_SPONTANEOUS) != 0)&&
         (hx_dfinfo[i].nb_of_callers == 0)&&
	 (strcmp(hx_dfinfo[i].name, "main") != 0)))
      {
      IFDECO
        printf("\"%s\" \"%s:%d\" \"%s\"\n", hx_dfinfo[i].name, ((hx_dfinfo[i].file!=NULL)?hx_dfinfo[i].file:"<n/a>"),
	                       hx_dfinfo[i].line, hx_dfinfo[i].object);
      else
        printf("%s : %s:%d : %s\n", hx_dfinfo[i].name, ((hx_dfinfo[i].file!=NULL)?hx_dfinfo[i].file:"<n/a>"),
	                       hx_dfinfo[i].line, hx_dfinfo[i].object);
      }
    }
  IFDECO
    printf("<details\n");
  else
    printf("\n");

  return(1);
}


void hx_display_stack(int nb_stack, HX_MemStack *stack, int options)
{
  int l;

  if (nb_stack == 0)
    return;

  IFDECO
    printf("stack ");
  for(l=0; l<nb_stack; l++)
    {
    if (stack[l].addr == NULL)
      {
      IFDECO
        printf("\"*\" ");
      }
    else
      {
      IFDECO
        {
	if (options & FNCCHK_MEMORY_LINE)
	  printf("\"%s\" \"%s:%d\" \"%s\" ", HX_DNAME(stack[l].name), HX_DFILE(stack[l].file),
	                          stack[l].line, HX_DOBJ(stack[l].object));
	printf("\"%s\" ", HX_DNAME(stack[l].name));
	}
      else
	{
	if (l == 0)
	  printf("     stack: ");
	if (options & FNCCHK_MEMORY_LINE)
	{
	  if (l)
	    printf("            ");
	  printf("%20s : %s (%s:%d)\n", HX_DOBJ(stack[l].object),
	    HX_DNAME(stack[l].name), HX_DFILE(stack[l].file), stack[l].line);
	}else
	{
	  if (l)
	    printf(" <- ");
  	  printf("%s", HX_DNAME(stack[l].name));
	  if (l+1 == nb_stack || stack[l+1].addr == NULL)
	    printf("\n");
	}
	}
      }
    }
  IFDECO
    printf("\n");
}



/* sub-functions used by memory_leaks and memory_blocks */
void hx_display_valid_block(HX_Mem *blk, int options)
{
  int j;
  HX_Mem *m;

  /* allocation part */
  IFDECO
    {
    printf("%p %u \"%s\" \"%s:%d\"", blk->ptr, blk->size_a,
	       HX_DNAME(blk->func_a), HX_DFILE(blk->file_a), blk->line_a);
    }
  else
    {
    printf("Block %d\n", blk->block);
    printf("   alloc, size: %p,%-8u ", blk->ptr, blk->size_a);
    printf("%s (%s:%d)\n",
	       HX_DNAME(blk->func_a), HX_DFILE(blk->file_a), blk->line_a);
    }
  for (j=0, m=blk->rnext; m; m=m->rnext,j++);
  IFDECO
    printf(" %d\n", j);
  /* reallocation part */
  for (j=0, m=blk->rnext; m; m=m->rnext,j++)
    {
    IFDECO
      {
      printf("%p %u \"%s\" \"%s:%d\" ", m->ptr, m->size_a,
		   HX_DNAME(m->func_a), HX_DFILE(m->file_a), m->line_a);
      }
    else
      {
      printf("  realloc(%3d): %p,%-8u ", j+1, m->ptr, m->size_a);
      printf("%s (%s:%d)\n", HX_DNAME(m->func_a),
                   HX_DFILE(m->file_a), m->line_a);
      }
    if (!m->rnext)
    {
      j++;
      break;
    }
    }
  if (j > 0)
    IFDECO
      printf("\n");

  /* free part */
  if (!m)
    m = blk;
  if (m->where_f == NULL)
    {
    IFDECO
      printf("\"*\" \"*\"\n");
    else
      printf("     never freed.\n");
    }
  else
    {
    IFDECO
      printf("\"%s\" \"%s:%d\"\n",HX_DNAME(m->func_f),
                            HX_DFILE(m->file_f), m->line_f);
    else
      printf("      freed at:                    %s (%s:%d)\n",
      			    HX_DNAME(m->func_f),
                            HX_DFILE(m->file_f), m->line_f);
    }
}
void hx_display_invalid_block(HX_Mem *blk, int options)
{
  if (blk->anext && blk->anext->where_f)
    {
    IFDECO
      {
      printf("still %p \"%s\" \"%s:%d\" \"%s\" \"%s:%d\"\n",
	 blk->ptr, HX_DNAME(blk->func_f),
	 HX_DFILE(blk->file_f), blk->line_f,
	 HX_DNAME(blk->anext->func_f),
	 HX_DFILE(blk->anext->file_f),
	 blk->anext->line_f);
      }
    else
      printf("Manipulation of address %p at %s (%s:%d),"
	     "    which was freed at %s (%s)\n",
	 blk->ptr, HX_DNAME(blk->func_f),
	 HX_DFILE(blk->file_f), blk->line_f,
	 HX_DNAME(blk->anext->func_f),
	 HX_DFILE(blk->anext->file_f),
	 blk->anext->line_f);
    }
  else
  if (blk->func_a == NULL)
    {/* invalid free */
    IFDECO
      printf("free %p \"%s\" \"%s:%d\"\n",
		blk->ptr, HX_DNAME(blk->func_f),
        	HX_DFILE( blk->file_f), blk->line_f);
    else
      printf("   unreferenced address (%p) for 'free' at %s (%s:%d)\n",
		blk->ptr, HX_DNAME(blk->func_f),
        	HX_DFILE( blk->file_f), blk->line_f);
    }
  else
    {/* invalid realloc */
    IFDECO
      printf("realloc %p \"%s\" \"%s:%d\"\n", 
		blk->ptr, HX_DNAME(blk->func_f),
        	HX_DFILE(blk->file_f), blk->line_f);
    else
      printf("   unreferenced address (%p) for 'realloc' at %s (%s:%d)\n",
		blk->ptr, HX_DNAME(blk->func_f),
        	HX_DFILE(blk->file_f), blk->line_f);
    }
}
void hx_display_valid_leak(HX_Mem *blk, int options)
{
  int j;
  HX_Mem *m;

  for (j=0,m=blk->rnext;m;m=m->rnext,j++);
  /* normal behavior */
  IFDECO
    {
    printf("%p %u \"%s\" \"%s:%d\" %d\n",blk->ptr, blk->size_a,
	     HX_DNAME(blk->func_a), HX_DFILE(blk->file_a), blk->line_a,
	     j);
    }
  else
    {
    printf("   Leak, size: %p,%-8u ", blk->ptr, blk->size_a);
    printf(" %s (%s:%d)\n", HX_DNAME(blk->func_a),
                                       HX_DFILE(blk->file_a), blk->line_a);
    }
  /*call-stack */
  hx_display_stack(blk->nb_stack_a, blk->stack_a, options);
  for (j=0,m=blk->rnext;m;m=m->rnext,j++)
    {
    IFDECO
      {
      printf("%p %u \"%s\" \"%s:%d\"\n", m->ptr, m->size_a,
          HX_DNAME(m->func_a),HX_DFILE(m->file_a),m->line_a);
      }
    else
      {
      printf("   realloc(%3d): %p,%-8u ", j+1, m->ptr, m->size_a);
      printf(" %s (%s:%d)\n",
          HX_DNAME(m->func_a),HX_DFILE(m->file_a),m->line_a);
      }
    /* new: call-stack */
    hx_display_stack(m->nb_stack_a, m->stack_a, options);
    }
  if (j > 0)
    IFDECO
      printf("\n");
}

static int hx_found_leak;

int hx_display_leak1(HX_Mem *me, int options)
{
  for (;me;me=me->anext)
  {
  if ((options & (FNCCHK_NO_UNRES_MALLOC|FNCCHK_NO_UNRES_REALLOC))&&
  	(me->where_a) && (!me->valid_a))
    continue;
  if ((options & FNCCHK_NO_UNRES_FREE)&&
  	(me->where_f) && (!me->valid_f))
    continue;

  if (me->where_a)
  {
    if (!me->where_f && !me->rnext)
    {
      if (!hx_found_leak)
      {
        hx_found_leak = 1;
        IFNDECO
        {
          printf("\n");
          printf(HX_TTL_MEM2);
          printf("\n");
        }
      }
      hx_display_valid_leak(me, options);
    }
  }
  else
  {
    if (!hx_found_leak)
    {
      hx_found_leak = 1;
      IFNDECO
      {
        printf("\n");
        printf(HX_TTL_MEM2);
        printf("\n");
      }
    }
    hx_display_invalid_block(me, options);
  }
  }
  return 0;
}

int hx_display_leaks(int options)
{
  if (!(options & FNCCHK_MEMORY))
    return(1);

  hx_found_leak = 0;

  if (!hx_mems)
    {
    IFDECO
      {
      printf(">leaks\n");
      printf("<leaks\n");
      }
    else
      printf("\nNo memory leaks (no memory block referenced).\n\n");
    return(1);
    }
  IFDECO
    printf(">leaks\n");

  avl_apply(hx_mems, (AVL_APPLY)hx_display_leak1, (void *)options, -1, AVL_INORDER);
  IFNDECO
    {
    if (!hx_found_leak)
      {
      printf("\nNo memory leaks.\n\n");
      }
    }
  IFDECO
    printf("<leaks\n");

  return(1);
}

int hx_display_mem1(HX_Mem *me, int options)
{
  for (;me;me=me->anext)
  {
  if ((options & (FNCCHK_NO_UNRES_MALLOC|FNCCHK_NO_UNRES_REALLOC))&&
  	(me->where_a) && (!me->valid_a))
    continue;
  if ((options & FNCCHK_NO_UNRES_FREE)&&
  	(me->where_f) && (!me->valid_f))
    continue;

  /* This is a realloc, someone else will print it */
  if (me->rprev)
    continue;

  if (me->where_a)
    hx_display_valid_block(me, options);
  else
    hx_display_invalid_block(me, options);
  }
  return 0;
}

int hx_display_memory(int options)
{
  if (!(options & FNCCHK_MEMORY_MORE))
    return(1);

  /** if requested, display memory usage details **/
  IFDECO
    {
    printf(">memory\n");
    }
  else
    {
    if (hx_mems)
      {
      printf("\n");
      printf(HX_TTL_MEM);
      printf("\n");
      }
    }
  if (hx_mems)
    avl_apply(hx_mems, (AVL_APPLY)hx_display_mem1, (void *)options, -1, AVL_INORDER);
  else
  IFNDECO
    printf("\nNo memory block referenced.\n\n");
  IFDECO
    printf("<memory\n");
  return(1);
}

#undef IFDECO
#undef IFNDECO

/* do the output of file stat. exec in the exec name,
   fstat is the file stat name, sort_by is the kind of sort,
   options is ORed options, fonly is NULL or a comma separated
   list of functions name to use, fnot is NULL or a comma separated
   list of functions to reject */
int fnccheck_dump(char *exec, char *fstat, char *objlist, int sort_by,
                  int options, char *fonly_, char *fnot_)
{
  char buffer2[HX_MAX_BUFFER]; /* temp buffer */
  void *fnc;
  FILE *sf;
  unsigned int i, j, k, nbinfo, nb_reallocs;
  int nbe, lng;
  unsigned long int itemp2;
  unsigned int ssu, nf;
  int nb_hide, nb_spont;
  double total;
  HX_Node **nodes=NULL;
  char *fonly=NULL, *fnot=NULL;
  char **fo_lst=NULL, **fn_lst=NULL;
  int is_main, updt;  /* true if total time is from main */
  int is_memory;


  /* register kill signals to make emergency exit */
  signal(SIGHUP,  hx_damned);
  signal(SIGINT,  hx_damned);
  signal(SIGTERM, hx_damned);
  signal(SIGQUIT, hx_damned);
  signal(SIGPIPE, hx_damned);
  signal(SIGTERM, hx_damned);


  /** special **/
  if (exec == NULL)
    {/* just estimate evarage time per call */
    /* extracted in 'fncaverage'. May not occur again */
    hxmessage("called with 'exec==NULL'. This is no more suported.");
    hxmessage("see 'fncaverage' instead.");
    return(1);
    }

  if (!(options & FNCCHK_CALLGRAPH))
    {
    if (!(options & FNCCHK_NO_HEADER))
      {
      if (options & FNCCHK_NO_DECORATION)
	{
	printf(">header\n");
	printf("%s %s %s\n", hx_my_name, FNCCHK_FULL_VERSION, FNCCHK_HEADERVERSION);
	printf("<header\n");
	}
      else
	{
	printf("\n");
	printf("FunctionChecker V%s for gcc (by H. Chu)\n", FNCCHK_FULL_VERSION);
	printf("Running profile data file version number %s\n", FNCCHK_HEADERVERSION);
	printf("http://sourceforge.net/projects/fnccheck/\n\n");
	}
      }
    }
  /** read the stat file **/
  hx_set_message("reading profile stat file");
  /* open the stat file */
  if ((sf = fopen(fstat==NULL?HX_DROP_FILE:fstat, "r")) == NULL)
    {
    hxmessage("can't open stat file '%s'.", fstat==NULL?HX_DROP_FILE:fstat);
    return(0);
    }
  /* header */
  fscanf(sf, "%1023s", buffer2);
  if (strcmp(buffer2, FNCCHK_HEADER) != 0)
    {
    hxmessage("invalid header file in '%s'.", fstat==NULL?HX_DROP_FILE:fstat);
    fclose(sf);
    return(0);
    }
  /* # of elements */
  fscanf(sf, "%u", &nbinfo);
  if (nbinfo <= 0)
    {
    hxmessage("invalid number of elements (%u)!", nbinfo);
    fclose(sf);
    return(0);
    }
printf("%u el\n", nbinfo);
  hx_nb_dfinfo = nbinfo;
  hx_allocate_fatal(hx_dfinfo, nbinfo);
  /* loop to read elements */
  for(i=0; i<nbinfo; i++)
    {
    hx_dfinfo[i].my_index   = i;
    hx_dfinfo[i].name       = NULL;
    hx_dfinfo[i].file	    = NULL;
    hx_dfinfo[i].line       = 0;
    hx_dfinfo[i].total      = 0.;
    hx_dfinfo[i].local      = 0.;
    hx_dfinfo[i].min        = 0.;
    hx_dfinfo[i].max        = 0.;
    
    fscanf(sf, "%p", &fnc);
    hx_dfinfo[i].function   = fnc;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].nb_calls   = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].nb_sec     = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].nb_usec    = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].nb_locsec  = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].nb_locusec = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].min_sec  = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].min_usec = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].max_sec  = (unsigned long int)itemp2;
    fscanf(sf, "%lu", &itemp2);
    hx_dfinfo[i].max_usec = (unsigned long int)itemp2;
    }
  fscanf(sf, "%u", &nb_reallocs);
  fscanf(sf, "%u", &ssu);
  fscanf(sf, "%u", &nf);
  for(i=0; i<nbinfo; i++)
    {
    fscanf(sf, "%u", &(hx_dfinfo[i].nb_of_callers));
    for(j=0; j<hx_dfinfo[i].nb_of_callers; j++)
      {
      fscanf(sf, "%p", &(hx_dfinfo[i].called_by[j]));
      fscanf(sf, "%p", &(hx_dfinfo[i].rcb[j].addr));
      fscanf(sf, "%x", &(hx_dfinfo[i].nb_called_by[j]));
      }
    }
  fscanf(sf, "%d", &hx_time_type);
  fscanf(sf, "%d", &is_memory);
  /* bogus line for executable path */
  fscanf(sf, "%1023s\n", buffer2);
  while (!feof(sf))
    {
      HX_DynObj *obj;
      void *ptr;
      fscanf(sf, "%p %1023s\n", &ptr, buffer2);
      obj = (HX_DynObj *)malloc(sizeof(HX_DynObj)+strlen(buffer2));
      obj->base = ptr;
      strcpy(obj->path, buffer2);
      hx_objects = realloc(hx_objects, (hx_nb_objects+1)*sizeof(HX_DynObj*));
      hx_objects[hx_nb_objects++] = obj;
    }
  fclose(sf);
printf("lecture fichier termine\n");
fflush(stdout);

  /** if requested, read the memory file **/
  if (options & (FNCCHK_MEMORY|FNCCHK_MEMORY_MORE))
    {
    if (is_memory)
      {
      hx_set_message("reading memory stat file");
      hx_compute_new_name(buffer2, (fstat==NULL?HX_DROP_FILE:fstat));
      if (!hx_read_memory_trace(options, buffer2))
	{
	hxmessage("error while reading memory file. Ignored.");
	}
      }
    else
      {
      hxmessage("memory profile requested, but corresponding profile");
      hxmessage("is not generated with memory profile enabled. Ignored.\n");
      }
    }

printf("names...\n");
fflush(stdout);
  hx_set_message("extracting symbol names");
  /** read symbols for the exec **/
  if (!hx_extract_names(options, exec, nbinfo))
    {
    hxmessage("error while reading functions name.");
    }

  /** parse elements to compute needed values **/
  for(i=0; i<nbinfo; i++)
    {
    hx_dfinfo[i].local = (double)hx_dfinfo[i].nb_locsec+(double)hx_dfinfo[i].nb_locusec/1000000.;
    hx_dfinfo[i].total = (double)hx_dfinfo[i].nb_sec+(double)hx_dfinfo[i].nb_usec/1000000.;
    hx_dfinfo[i].min   = (double)hx_dfinfo[i].min_sec+(double)hx_dfinfo[i].min_usec/1000000.;
    hx_dfinfo[i].max   = (double)hx_dfinfo[i].max_sec+(double)hx_dfinfo[i].max_usec/1000000.;
    }
printf("names ok\n");
printf("preparation call-graph\n");
fflush(stdout);
  hx_set_message("building call-graph");
  /** now construct the call-graph of the profile **/
  if (fonly_ != NULL)
    fonly = strdup(fonly_);
  if (fnot_ != NULL)
    fnot  = strdup(fnot_);
  /* create the list of nodes */
  hx_allocate_fatal(nodes, nbinfo);
  for(i=0; i<nbinfo; i++)
    {
    if ((nodes[i] = hx_create_node(&(hx_dfinfo[i]))) == NULL)
      {
      hxfatal_msg();
      return(0);
      }
    }
printf("nodes ok\n");
fflush(stdout);
  /* now add the links between nodes */
  for(i=0; i<nbinfo; i++)
    {
    for(j=0; j<hx_dfinfo[i].nb_of_callers; j++)
      {
      hx_add_child(nodes[hx_get_func_by_ptr(hx_dfinfo[i].called_by[j], nbinfo)-1], nodes[i]);
      hx_add_prev(nodes[i], nodes[hx_get_func_by_ptr(hx_dfinfo[i].called_by[j], nbinfo)-1]);
      }
    }
printf("liens cg ok\n");
fflush(stdout);
  /* search if main is here */
  is_main = -1;
  for(i=0; i<nbinfo; i++)
    {
    if (hx_dfinfo[i].name != NULL)
      if (strcmp(hx_dfinfo[i].name, "main") == 0)
        {
        is_main = i;
        break;
        }
    }
  /* the sum of local times (but it is < at the real total time) */
  total = 0.;
  for(i=0; i<nbinfo; i++)
    {
    total += hx_dfinfo[i].local;
    }
  /* now look if a function got a total time > to this */
  updt = 0;
  for(i=0; i<nbinfo; i++)
    {
    if (hx_dfinfo[i].total > total)  /* it is better */
      {
      total = hx_dfinfo[i].total;
      updt = 1;
      }
    }
  if (is_main < 0)
    is_main = 0;  /* no main */
  else
    {
    /* is it time from main ? */
    if (ABS(total-hx_dfinfo[is_main].total) < EPS)
      is_main = 1;    /* not realy precise */
    }
  if (total < EPS)
    {/* to prevent /0 error */
    total = 0.;
    }
  else
    {
    /* add a small amount of time. It can cause max time not
       to be 100% but the alternative is a time > to 100%!.
       This is due to the fact the the sum of local times is
       a little bit small that total time (caused by the
       calling time to treatment functions) */
     if (!updt)
      total += total/100.;  /* add 1% */
    }
printf("call-graph et temps max ok\n");
fflush(stdout);

  /* display the cycles list (just displayed, not stored) */
  if (options & FNCCHK_CYCLES)
    {
    if (options & FNCCHK_NO_DECORATION)
      printf(">cycle\n");
    else
      printf("Cycles during execution:\n\n");
    if (!hx_compute_cycles(options, nodes, nbinfo))
      {
      hxmessage("error while computing cycles.");
      }
    if (options & FNCCHK_NO_DECORATION)
      printf("<cycle\n");
    else
      printf("\n");
    }

printf("propag restrictions\n");
fflush(stdout);
  hx_set_message("propagating function restrictions");
  /* extract functions name from lists */
  /* if both are used, 'fnot' is performed first,
     and then 'fonly' is added, in order to
     re-set functions which are shared by 'fnot's */ 
  nbe = 0;
  if (fonly != NULL)
    {
    for(i=0; i<strlen(fonly); i++)
      {
      if (fonly[i] == ',')
        nbe++;
      }
    hx_allocate_fatal(fo_lst, (nbe+2));
    fo_lst[nbe+1] = NULL;
    fo_lst[0] = &(fonly[0]);
    j = 1;
    lng = strlen(fonly);
    for(i=1; i<lng; i++)
      {
      if (fonly[i] == ',')
        {
	fo_lst[j] = &(fonly[i+1]);
	fonly[i] = '\0';
        j++;
	}
      }
    }
  if (fnot != NULL)
    {
    for(i=0; i<strlen(fnot); i++)
      {
      if (fnot[i] == ',')
        nbe++;
      }
    hx_allocate_fatal(fn_lst, (nbe+2));
    fn_lst[nbe+1] = NULL;
    fn_lst[0] = &(fnot[0]);
    j = 1;
    lng = strlen(fnot);
    for(i=1; i<lng; i++)
      {
      if (fnot[i] == ',')
        {
	fn_lst[j] = &(fnot[i+1]);
	fnot[i] = '\0';
        j++;
	}
      }
    }
  /* for each function, set it to "used" or "removed" */
  for(i=0; i<nbinfo; i++)
    {
    nodes[i]->treated = 0;
    if (fnot != NULL)
      {
      if (hx_is_in_list(nodes[i]->function->name, fn_lst))
        {
	nodes[i]->keep = 0;
	}
      else
        {
	nodes[i]->keep = 1;
	}
      }
    else
    if (fonly != NULL)
      {
      if (hx_is_in_list(nodes[i]->function->name, fo_lst))
        {
	nodes[i]->keep = 1;
	}
      else
        {
	nodes[i]->keep = 0;
	}
      }
    }
  /* propagate "used" or "removed" in call-graph if requested */
  if (options & FNCCHK_USE_CHILDS)
    {
    if (fnot != NULL)
      {
      for (i=0; i<nbinfo; i++)
        {
	if (nodes[i]->keep == 0)
	  {
	  nodes[i]->keep = 1;
          hx_propagate_to_child(nodes[i], 0);
	  }
	}
      }
    if ((fonly != NULL)&&(fnot != NULL))
      {/* just activate needed ones */
      for(i=0; i<nbinfo; i++)
        nodes[i]->treated = 0;
      for(i=0; i<nbinfo; i++)
        {
        if (hx_is_in_list(nodes[i]->function->name, fo_lst))
	  {
          hx_propagate_to_child_p(nodes[i]);
	  }
	}
      }
    else
    if (fonly != NULL)
      {
      for (i=0; i<nbinfo; i++)
        {
	if (nodes[i]->keep == 1)
	  {
	  nodes[i]->keep = 0;
          hx_propagate_to_child(nodes[i], 1);
	  }
	}
      }
    }

  /* propagate "used" or "removed" in call-graph if requested */
  if (options & FNCCHK_USE_CALLERS)
    {
    if (fnot != NULL)
      {
      for (i=0; i<nbinfo; i++)
        {
	if (nodes[i]->keep == 0)
	  {
	  nodes[i]->keep = 1;
          hx_propagate_to_caller(nodes[i], 0);
	  }
	}
      }
    if ((fonly != NULL)&&(fnot != NULL))
      {/* just activate needed ones */
      for(i=0; i<nbinfo; i++)
        nodes[i]->treated = 0;
      for(i=0; i<nbinfo; i++)
        {
        if (hx_is_in_list(nodes[i]->function->name, fo_lst))
	  {
          hx_propagate_to_caller_p(nodes[i]);
	  }
	}
      }
    else
    if (fonly != NULL)
      {
      for (i=0; i<nbinfo; i++)
        {
	if (nodes[i]->keep == 1)
	  {
	  nodes[i]->keep = 0;
          hx_propagate_to_caller(nodes[i], 1);
	  }
	}
      }
    }
  
  
printf("liste finale\n");
fflush(stdout);
  /* create the final list */
  hx_tmp_dfinfo = hx_dfinfo;
  hx_nb_tmp_dfinfo = hx_nb_dfinfo;
  hx_nb_dfinfo = 0;
  /* how many entities */
  for(i=0; i<nbinfo; i++)
    {
    if (nodes[i]->keep)
      hx_nb_dfinfo++;
    }
  hx_allocate_fatal(hx_dfinfo, hx_nb_dfinfo);
  j = 0;
  for(i=0; i<nbinfo; i++)
    {
    if (nodes[i]->keep)
      {
      memcpy((void*)&(hx_dfinfo[j]), nodes[i]->function, sizeof(HX_FDInfo));
      j++;
      }
    }
  nbinfo = hx_nb_dfinfo;  /* new list completed */
  /* cleanup callers (remove those which are not in final list) */
  for(i=0; i<nbinfo; i++)
    {
    k = 0;
    for(j=0; j<hx_dfinfo[i].nb_of_callers; j++)
      {
      hx_dfinfo[i].called_by[k] = hx_dfinfo[i].called_by[j];
      if (hx_get_func_by_ptr(hx_dfinfo[i].called_by[j], nbinfo))
        {/* keep the caller */
	k++;
	}
      }
    hx_dfinfo[i].nb_of_callers = k;
    }


printf("max (bis)\n");
fflush(stdout);
  hx_set_message("computing total execution time");
  /* if 'effective max time', recompute total time using this list */
  if (options & FNCCHK_MAXTIME_EFF)
    {
    /* search for total time: 'main' time if detected, else
       we compute the sum of each local time */
    total = 0.;
    is_main = 0;
    for(i=0; i<nbinfo; i++)
      {
      if (strcmp(hx_dfinfo[i].name, "main") == 0)
        {
	total = hx_dfinfo[i].total;
	is_main = 1;
	break;
	}
      }
    if (total == 0.)
      {
      /* the sum */
      total = 0.;
      for(i=0; i<nbinfo; i++)
	{
	total += hx_dfinfo[i].local;
	}
      /* add a small amount of time. It can cause max time not
	 to be 100% but the alternative is a time > to 100%!.
	 This is due to the fact the the sum of local times is
	 a little bit small that total time (caused by the
	 calling time to treatment functions) */
      total += total/100.;  /* add 1% */
      }
    if (total < EPS)
      {/* total=0 => no percentages performed */
      total = 0.;
      }
    }

  /** if requested, output call-graph and exit **/
  if (options & FNCCHK_CALLGRAPH)
    {
    hx_set_message("displaying call-graph (VCG)");
    printf("graph: {\n");
    printf(" orientation: left_to_right\n");
    /* display nodes */
    for(i=0; i<nbinfo; i++)
      {
      printf("  node: { title: \"%d\" label: \"%s\" borderwidth:0}\n",
                 i, hx_dfinfo[i].name);
      }
    /* display edges */
    for(i=0; i<nbinfo; i++)
      {
      j = 0;
      while(nodes[i]->nexts[j] != NULL)
        {
        printf("  edge: { sourcename: \"%d\" targetname: \"%d\" thickness: 1}\n",
                    i, nodes[i]->nexts[j]->function->my_index);
	j++;
	}
      }
    printf("    }\n");
    
    /** freeing all datas **/
    for(i=0; i<nbinfo; i++)
      {
      /* NOT ACTIVATED: name computation have changed...
      if (hx_dfinfo[i].name != NULL)
	free(hx_dfinfo[i].name);
      */
      }
    free(hx_dfinfo);
    return(1);
    }

printf("tri\n");
fflush(stdout);
  hx_set_message("sorting data");
  /** sort the table **/
  if (sort_by < 0)
    {
    hx_mult_for_sort *= -1;
    sort_by = -sort_by;
    }
  switch(sort_by)
    {
    case FNCCHK_SORT_NO:  /* no sort */
      break;
    case FNCCHK_SORT_LOCAL:
      qsort((void*)hx_dfinfo, (size_t)nbinfo, sizeof(HX_FDInfo), hx_sort_by_local);
      break;
    default:   /* tri par defaut */
    case FNCCHK_SORT_TOTAL:
      qsort((void*)hx_dfinfo, (size_t)nbinfo, sizeof(HX_FDInfo), hx_sort_by_total);
      break;
    case FNCCHK_SORT_NAME:
      qsort((void*)hx_dfinfo, (size_t)nbinfo, sizeof(HX_FDInfo), hx_sort_by_name);
      break;
    case FNCCHK_SORT_CALLS:
      qsort((void*)hx_dfinfo, (size_t)nbinfo, sizeof(HX_FDInfo), hx_sort_by_calls);
      break;
    }

  nb_hide = 0;
  nb_spont = 0;
  for(i=0; i<nbinfo; i++)
    {
    if (((options & FNCCHK_NO_UNDEF) != 0)&&(!hx_dfinfo[i].valid))
      {
      nb_hide++;
      }
    else
      {
      if (((options & FNCCHK_NO_SPONTANEOUS) != 0)&&
          (hx_dfinfo[i].nb_of_callers == 0)&&
	  (strcmp(hx_dfinfo[i].name, "main") != 0))
        {
	nb_spont++;
	}
      }
    }

  hx_set_message("displaying general informations");
  /** generate formated output for the stats **/
  if (!(options & FNCCHK_NO_GENERAL))
    {
    if (options & FNCCHK_NO_DECORATION)
      {
      printf(">general\n");
      printf("\"%s\"\n", exec);
      printf("%f", total);
      if (!is_main)
	printf(" 2\n");
      else
	printf(" 1\n");
      if (hx_time_type == HX_TIME_EXT)
	{
	printf("1\n");
	}
      else
      if (hx_time_type == HX_TIME_CPU)
	{
	printf("2\n");
	}
      else
      if (hx_time_type == HX_TIME_SYS)
	{
	printf("3\n");
	}
      else
	{
	printf("0\n");
	}
      printf("%d\n", nb_reallocs);
      printf("%d\n", nb_hide);
      printf("%d\n", nb_spont);
      printf("%d\n", hx_nb_tmp_dfinfo-nbinfo);
      if (is_memory)
        {
	printf("%ld %ld %ld %f %f\n", hx_nb_mem_max, hx_nb_heap_size,
	           hx_nb_mem_used,
		   ((hx_nb_mem_max==0)?0.:100.0*hx_nb_mem_used/hx_nb_mem_max),
		   ((hx_nb_heap_size==0)?0.:100.0*hx_nb_mem_used/hx_nb_heap_size));
	}
      printf("%u\n", ssu);
      printf("%u\n", nf);
      printf("<general\n");
      }
    else
      {
      printf("General information:\n\n");
      printf("Profile for '%s'\n", exec);
      printf("Total execution time: %f", total);
      if (!is_main)
	printf(" (not main() time)");
      printf("\n");
      printf("Times computed using ");
      if (hx_time_type == HX_TIME_EXT)
	{
	printf("real clock time.\n");
	}
      else
      if (hx_time_type == HX_TIME_CPU)
	{
	printf("CPU process clock time.\n");
	}
      else
      if (hx_time_type == HX_TIME_SYS)
	{
	printf("CPU process plus SYSTEM clock time.\n");
	}
      else
	{
	printf("unknown clock time (invalid argument in stat. file!).\n");
	}
      printf("Number of realloc performed: %d\n", nb_reallocs);
      printf("Unresolved functions not shown    : %d\n", nb_hide);
      printf("Spontaneous functions not shown   : %d\n", nb_spont);
      printf("Hidden functions due to -not/-only: %d\n", hx_nb_tmp_dfinfo-nbinfo);
      if (is_memory)
        {
	printf("Max memory allocated            : %ld\n", hx_nb_mem_max);
	printf("Total heap memory used          : %ld\n", hx_nb_heap_size);
	printf("Total unfreed memory            : %ld  (%4.1f%% mem, %4.1f%% heap)\n",
	           hx_nb_mem_used,
		   ((hx_nb_mem_max==0)?0.:100.0*hx_nb_mem_used/hx_nb_mem_max),
		   ((hx_nb_heap_size==0)?0.:100.0*hx_nb_mem_used/hx_nb_heap_size));
	}
      printf("Final stack size used: %u\n", ssu);
      printf("Number of function(s): %u\n\n", nf);
      }
    }

  for (i=0; i<nbinfo; i++)
    if (hx_dfinfo[i].name == NULL)
      {
	char buf[40];
	sprintf(buf,"%#x", hx_dfinfo[i].function);
	hx_dfinfo[i].name = strdup(buf);
      }

  /* display flat profile */
  hx_set_message("displaying flat profile");
  hx_display_profile(options, nbinfo, total);

  /* display MIN/MAX times */
  hx_set_message("displaying MIN-MAX list");
  hx_display_minmax(options, nbinfo);

  /* display functions details (file:line) */
  hx_display_functions(options, nbinfo);

  /* display who calls who */
  hx_set_message("displaying call-graph");
  hx_display_calls(options, nbinfo, nb_hide, nb_spont);


  /* if memory profile anabled */
  if (is_memory)
    {
    /** if requested, display unfreed memory **/
    hx_set_message("displaying memory leaks");
    hx_display_leaks(options);

    /* display memory blocks */
    hx_set_message("displaying memory blocks");
    hx_display_memory(options);
    }

  /** freeing all datas **/
  hx_set_message("freeing data");
  for(i=0; i<nbinfo; i++)
    {
    if (hx_dfinfo[i].name != NULL)
      free(hx_dfinfo[i].name);
    }
  free(hx_dfinfo);
  if (nodes != NULL)
    free(nodes);

  /** all right, just leave **/
  hx_set_message("exiting");
  return(1);
}


/*
display explanations about what 'fnccheck_dump' print.
also display details on stat-file format
*/
int hx_details(char *name)
{
  /* header */
  printf("\n");
  printf("FunctionChecker for gcc (by Hexasoft)\n\n");
  printf("\n");

  printf("WARNING: this text is not up to date... see the 'fncdump.info'\n"
         "  file or the '*.html' files in 'doc' dir of 'fnccheck'.\n\n");

  /* displayed datas */
  printf("Total execution time: ???  (a)\n");
  printf("Times computed using ???.  (b)\n");
  printf("|%20s|%20s|%12s|%13s|\n", "local      ", "total      ", "", "6 ");
  printf("|%13s|%6s|%13s|%6s|%12s|%13s| 7 name\n\n", "sec.1", "% 2", "sec.3", "% 4", "calls  5 ", "tot. sec/call");

  printf("a. total execution time\n");
  printf("b. kind of time used: real, CPU or SYSTEM\n");
  printf("1. local time spend in function. local time is total time (3) minus\n"
         "      time spend in known child-functions\n");
  printf("4. percentage of (1) regards to total execution time\n");
  printf("3. total number of seconds between enters and exits for this function\n");
  printf("4. percentage of (3) regards to total execution time\n");
  printf("5. number of function calls\n");
  printf("6. average total time per call for function\n");
  printf("7. name of the function (or symbol if not available)\n");
  printf("\n");

  printf("Unresolved function(s) not shown  : ??  (1)\n"
         "Spontaneous function(s) not shown : ??  (2)\n"
         "Hidden functions due to -not/-only: ??  (3)\n"
         "Final stack size used: ???              (4)\n"
         "Number of function(s): ???              (5)\n\n");

  printf("1. number of unresolved function hidden (see -no-unresolved)\n");
  printf("2. number of spontaneous function hidden (see -no-spontaneous)\n");
  printf("3. number of hidden functions due to -not/-only (see -not/-only)\n");
  printf("4. final stack size used during execution\n");
  printf("5. total number of functions stored during execution\n");
  printf("\n");

  printf("|%13s|%13s| name  3\n", "MIN  1  ", "MAX   2 ");
  printf("|-------------|-------------|-----\n\n");

  printf("1. minimum (total) time spend in function\n");
  printf("2. maximum (total) time spend in function\n");
  printf("3. name of the function\n");
  printf("\n");

  printf("'xxx' [n] called by:  (OR 'xxx' [n] called spontaneously.)\n"
         "[n1], [n2] (...)\n"
	 "(...)\n\n");

  printf("List of functions indexes ([n1], [n2]...) which call 'xxx'\n");
  printf("'called spontaneously' indicate that no one calls it (i.e. main)\n");
  printf("Note: * with -calls option, the list of functions CALLED BY 'xxx'\n"
         "          is displayed here.\n");
  printf("      * with +calls option, both this list and the CALLED BY list\n"
         "          are displayed.\n");
  printf("      * option -call-details add the file and line where calls\n"
         "          happend.\n");
  printf("      * option -func-details displays the list of functions, with\n"
         "          their correponding file and line.\n");
  printf("\n");

  /* file format */
  printf("Stat-file format (%s):\n", FNCCHK_FULL_VERSION);
  printf("\n"
	" HEADER\n"
	" # of elements\n"
	" f1_ptr  nb_calls nb_sec nb_usec nb_locsec nb_locusec\n"
	"        min_sec min_usec max_sec max_usec\n"
	" f2_ptr  nb_calls nb_sec nb_usec nb_locsec nb_locusec\n"
	"        min_sec min_usec max_sec max_usec\n"
	" ...\n"
	" # of realloc\n"
	" stack_size func_table_size\n"
	" nb_of_callers_f1 caller1_ptr rc1_ptr caller2_ptr rc1_ptr ...\n"
	" nb_of_callers_f2 caller1_ptr rc2_ptr caller2_ptr rc2_ptr ...\n"
	" ...\n"
	" time_mode\n"
	" exec_name\n");
  printf("HEADER    : identifier (including file-format version) of stat-files\n");
  printf("# el      : number of functions\n");
  printf("fi_ptr    : symbol of the function\n");
  printf("nb_calls  : number of calls for this function\n");
  printf("nb_sec    : total time in the function\n");
  printf("nb_usec   : corresponding usec\n");
  printf("_locsec   : local time in the function\n");
  printf("_locusec  : corresponding usec\n");
  printf("min_sec   : minimum time spend in the function\n");
  printf("_subusec  : corresponding usec\n");
  printf("min_sec   : maximum time spend in the function\n");
  printf("_subusec  : corresponding usec\n");
  printf("#realloc  : number of 'realloc' performed during execution\n");
  printf("stack_size: final stack size used during execution\n");
  printf("func_tbl_s: number of function stored during execution\n");
  printf("nbofcaller: number of callers for the corresp. function\n");
  printf("calleri_pt: symbol of ith caller for the function\n");
  printf("rci_ptr   : symbol of caller of the function\n");
  printf("time_mode : the FNCCHK_TIME type\n");
  printf("exec_name : corresp. exec name (not implemented)\n");
  
  printf("\n");
  return(1);
}



/* usage. called by main() */
void usage(char *name)
{
  int i;

  printf("fncdump by Highland Sun (H.Chu, December 2001)\n");
  printf("Usage: %s exec [opts]\n", name);
  printf("   or: %s -avg\n", name);
  printf("Opts:  -sfile f        : use 'f' as stat file instead of 'fnccheck.out'\n");
  printf("       -sort n         : sort mode. See at the end for details\n");
  printf("       +sort n         : sort mode (reverse). See at the end for details\n");
  printf("       -no-spontaneous : dont print spontaneous called functions (dont apply to main)\n");
  printf("       -no-unresolved  : dont print unresolved symbols\n");
  printf("       -no-bad-malloc  : dont print memory blocks/leaks with unresolved malloc place\n");
  printf("       -no-bad-realloc : dont print memory blocks/leaks with unresolved realloc places\n");
  printf("       -no-bad-free    : dont print memory blocks/leaks with unresolved free place\n");
  printf("       -calls          : show 'called' functions instead of 'called by'\n");
  printf("       +calls          : show 'called' functions AND 'called by'\n");
  printf("       -no-minmax      : dont display MIN/MAX time for functions\n");
  printf("       -func-details   : add file/line for functions\n");
  printf("       -call-details   : add file/line for function calls\n");
  printf("       -fullname       : use full pathname for file/line info\n");
  printf("       -only <lst>     : just use functions that are in the comma separated list\n");
  printf("       -not <lst>      : dont use functions that are in the comma separated list\n");
  printf("       -propagate      : -only/-not is applied to childs\n");
  printf("       -rpropagate     : -only/-not is applied to callers\n");
  printf("       -real-maxtime   : use total execution time of displayed functions rather\n"
         "                         than total exec. time of all functions.\n");
  printf("       -call-graph     : output the call-graph in VCG format (nothing else displayed)\n");
  printf("       -cycles         : display detected cycles during execution\n");
  printf("       -memory         : display unfreed memory in program (see FNCCHK_MEMORY)\n");
  printf("       -no-decoration  : do not display decorations (for computer treatments)\n");
  printf("       -no-profile     : do not display flat profile\n");
  printf("       -no-calls       : do not display call-graph\n");
  printf("       -no-general     : do not display 'general' section (# of functions...)\n");
  printf("       -no-header      : do not display header (version, name...)\n");
  printf("       -cumul          : add a 'cumul' column in flat profile, with cumulated\n"
         "                         local times (in %%). Implies '-sort 1'\n");
  printf("       -memory-more    : display all memory activity in program (see FNCCHK_MEMORY)\n");
  printf("       -memory-stack-more: display file and line for call-stack in memory leaks\n");
  printf("    -sum-profile a b c : sum profile files 'a' and 'b' into 'c' ('-' means 'stdout')\n");
  printf("       --help          : this message\n");
  printf("       --version       : fnccheck/%s version\n", name);
  printf("       --misc          : author/contact/bugs\n");
  printf("       --details       : explanations about displayed informations\n");
  printf("(1): if 'fncdump' is compiled using 'libbfd' (standard behavior)\n");
  printf("     else (make fncdump_nobfd), -addr2line is the default approach\n");
  printf("Sort types:\n");
  for(i=FNCCHK_SORT_MIN; i<=FNCCHK_SORT_MAX; i++)
    printf("    %d:  sorted by '%s'\n", i, fncchk_sort_names[i]);
  printf("The -avg usage gives you the average time spend\n"
         "  in FncCheck treatments per each of your function.\n");
}

void misc(char *name)
{
  printf("FunctionCheck (%s) %s by Howard Chu (Highland Sun), December 2001.\n",
              name, FNCCHK_FULL_VERSION);
  printf("For gcc V2.95.2 and higher.\n");
  printf("Author:   H. Chu (Highland Sun)\n");
  printf("Original: Y. Perret (Hexasoft)\n");
  printf("Helper:   Petter Reinholdtsen (pere)\n");
  printf("EMail:    hyc@highlandsun.com\n");
  printf("HTTP:     http://freshmeat.net/projects/fncchk/\n");
  printf("Send bug reports using email.\n");
}


/* options list */
typedef struct
{
  int id;
  char *opts[8];  /* max 7 alternative names   */
  int sup;        /* because last must be NULL */
}HX_Option;
typedef enum
{
  opt_ndef=0,
  opt_help,
  opt_version,
  opt_misc,
  opt_details,
  opt_average,
  opt_propagate,
  opt_rpropagate,
  opt_nm,
  opt_calls,
  opt_pcalls,
  opt_sort,
  opt_usort,
  opt_only,
  opt_not,
  opt_sfile,
  opt_call_details,
  opt_func_details,
  opt_no_minmax,
  opt_no_unresolved,
  opt_no_spontaneous,
  opt_addr2line,
  opt_fullname,
  opt_maxtimeeff,
  opt_callgraph,
  opt_cycles,
  opt_sumprof,
  opt_memory,
  opt_memory_more,
  opt_no_deco,
  opt_no_profile,
  opt_no_call,
  opt_no_general,
  opt_no_header,
  opt_tmp_file,
  opt_cumul,
  opt_memory_line,
  opt_no_demangle,
  opt_no_ur_malloc,
  opt_no_ur_free,
  opt_no_ur_realloc,
  opt_pack_mem,
  opt_the_end,
}HXOL;
HX_Option hx_options[]=      {
    { opt_help, {"--help","-h","-help","--h",NULL}, 0},
    { opt_version, {"--version","-v","-version","--v",NULL}, 0},
    { opt_misc, {"--misc",NULL}, 0},
    { opt_details, {"--details",NULL}, 0},
    { opt_average, {"-avg","-average",NULL}, 0},
    { opt_propagate, {"-propagate",NULL}, 0},
    { opt_rpropagate, {"-rpropagate",NULL}, 0},
    { opt_no_spontaneous, {"-no-spontaneous",NULL}, 0},
    { opt_no_unresolved, {"-no-unresolved",NULL}, 0},
    { opt_no_minmax, {"-no-minmax",NULL}, 0},
    { opt_nm, {"-nm",NULL}, 0},
    { opt_calls, {"-calls",NULL}, 0},
    { opt_pcalls, {"+calls",NULL}, 0},
    { opt_func_details, {"-function-details","-func-details","-func+",NULL}, 0},   /* -func+ is here for compatibility */
    { opt_call_details, {"-call-details", "-calls+",NULL}, 0}, /* -calls+ is here for compatibility */
    { opt_sfile, {"-sfile", "-stat-file",NULL}, 1},
    { opt_sort, {"-sort",NULL}, 1},
    { opt_usort, {"+sort",NULL}, 1},
    { opt_only, {"-only",NULL}, 1},
    { opt_not, {"-not",NULL}, 1},
    { opt_addr2line, {"-addr2line",NULL}, 0},
    { opt_fullname, {"-fullname",NULL}, 0},
    { opt_maxtimeeff, {"-real-maxtime",NULL}, 0},
    { opt_callgraph, {"-call-graph",NULL}, 0},
    { opt_cycles, {"-cycles",NULL}, 0},
    { opt_sumprof, {"-sum-profiles","-sum-profile",NULL}, 3},
    { opt_memory, {"-memory",NULL}, 0},
    { opt_memory_more, {"-memory-more",NULL}, 0},
    { opt_no_deco, {"-no-decoration","-no-decorations",NULL}, 0},
    { opt_no_profile, {"-no-profile","-no-flat-profile",NULL}, 0},
    { opt_no_call, {"-no-call","-no-calls","-no-graph-call",NULL}, 0},
    { opt_no_general, {"-no-general",NULL}, 0},
    { opt_no_header, {"-no-header",NULL}, 0},
    { opt_tmp_file, {"-tmp-dir","-temp-dir",NULL}, 1},
    { opt_cumul, {"-cumul",NULL}, 0},
    { opt_memory_line, {"-memory-stack-more",NULL}, 0},
    { opt_no_demangle, {"-no-demangle",NULL}, 0},
    { opt_no_ur_malloc, {"-no-bad-malloc",NULL}, 0},
    { opt_no_ur_free, {"-no-bad-free",NULL}, 0},
    { opt_no_ur_realloc, {"-no-bad-realloc",NULL}, 0},
    { opt_pack_mem, {"-pack-mallocs","-pack-malloc",NULL}, 0},
    { opt_ndef, {NULL}, 0},
                            };

/* test if a given option is present */
int hx_test_option(HXOL opt, int i, int argc, char* argv[])
{
  int l, el;

  /* valid ? */
  el = 0;
  while(hx_options[el].id != opt_ndef)
    {
    if (hx_options[el].id == opt)
      break;
    el++;
    }
  if (hx_options[el].id == opt_ndef)
    return(0);   /* no match! */
  /* bounds */
  if (i+hx_options[el].sup >= argc)
    return(0);
  l = 0;
  while(hx_options[el].opts[l] != NULL)
    {
    if (strcmp(hx_options[el].opts[l], argv[i]) == 0)
      return(1);
    l++;
    }
  return(0);
}

/* just call the dump function on the given prog,
   with the alternate stat file "-file f" and the sort mode
   "-sort n", plus options and functions list */
int main(int argc, char *argv[])
{
  char *filename=NULL, *statname=NULL;
  char *objlist=NULL;
  int mode=0, options=FNCCHK_NONE;
  char *fonly=NULL, *fnot=NULL;
  int i;

  if (argc < 2)
    {
    usage(argv[0]);
    return(0);
    }


#define HX_TEST_OPTION(n) hx_test_option(n, i, argc, argv)
  /* parse options */
  for(i=1; i<argc; i++)
    {
    if ((argv[i][0] != '-')&&(argv[i][0] != '+'))
      {
      if ((filename = strdup(argv[i])) == NULL)
	{
	fprintf(stderr, "%s: Allocation error.\n", argv[0]);
	return(9);
	}
      }
    else
    if (HX_TEST_OPTION(opt_average))
      {/* special */
      if (!system("fncaverage"))
	{
	fprintf(stderr, "%s: Error while computing average time.\n", argv[0]);
	return(3);
	}
      return(0);
      }
    else
    if (HX_TEST_OPTION(opt_help))
      {
      usage(argv[0]);
      return(0);
      }
    else
    if (HX_TEST_OPTION(opt_misc))
      {
      misc(argv[0]);
      return(0);
      }
    else
    if (HX_TEST_OPTION(opt_details))
      {
      hx_details(argv[0]);
      return(0);
      }
    else
    if (HX_TEST_OPTION(opt_version))
      {
      printf("%s\n", FNCCHK_VERSION);
      return(0);
      }
    else
    if (HX_TEST_OPTION(opt_propagate))
      {
      options |= FNCCHK_USE_CHILDS;
      }
    else
    if (HX_TEST_OPTION(opt_rpropagate))
      {
      options |= FNCCHK_USE_CALLERS;
      }
    else
    if (HX_TEST_OPTION(opt_no_spontaneous))
      {
      options |= FNCCHK_NO_SPONTANEOUS;
      }
    else
    if (HX_TEST_OPTION(opt_maxtimeeff))
      {
      options |= FNCCHK_MAXTIME_EFF;
      }
    else
    if (HX_TEST_OPTION(opt_callgraph))
      {
      options |= FNCCHK_CALLGRAPH;
      }
    else
    if (HX_TEST_OPTION(opt_cumul))
      {
      options |= FNCCHK_CUMUL;
      mode = FNCCHK_SORT_LOCAL;
      }
    else
    if (HX_TEST_OPTION(opt_memory_line))
      {
      options |= FNCCHK_MEMORY_LINE;
      }
    else
    if (HX_TEST_OPTION(opt_no_demangle))
      {
      options |= FNCCHK_NO_DEMANGLE;
      }
    else
    if (HX_TEST_OPTION(opt_cycles))
      {
      options |= FNCCHK_CYCLES;
      }
    else
    if (HX_TEST_OPTION(opt_no_ur_malloc))
      {
      options |= FNCCHK_NO_UNRES_MALLOC;
      }
    else
    if (HX_TEST_OPTION(opt_no_ur_free))
      {
      options |= FNCCHK_NO_UNRES_FREE;
      }
    else
    if (HX_TEST_OPTION(opt_no_ur_realloc))
      {
      options |= FNCCHK_NO_UNRES_REALLOC;
      }
    else
    if (HX_TEST_OPTION(opt_pack_mem))
      {
      options |= FNCCHK_PACK_MEMORY;
      }
    else
    if (HX_TEST_OPTION(opt_no_unresolved))
      {
      options |= FNCCHK_NO_UNDEF;
      }
    else
    if (HX_TEST_OPTION(opt_no_profile))
      {
      options |= FNCCHK_NO_PROFILE;
      }
    else
    if (HX_TEST_OPTION(opt_no_header))
      {
      options |= FNCCHK_NO_HEADER;
      }
    else
    if (HX_TEST_OPTION(opt_no_deco))
      {
      options |= FNCCHK_NO_DECORATION;
      }
    else
    if (HX_TEST_OPTION(opt_no_general))
      {
      options |= FNCCHK_NO_GENERAL;
      }
    else
    if (HX_TEST_OPTION(opt_no_call))
      {
      options |= FNCCHK_NO_CALL;
      }
    else
    if (HX_TEST_OPTION(opt_calls))
      {
      options |= FNCCHK_CALLS;
      }
    else
    if (HX_TEST_OPTION(opt_no_minmax))
      {
      options |= FNCCHK_NO_MINMAX;
      }
    else
    if (HX_TEST_OPTION(opt_func_details))
      {
      options |= FNCCHK_FUNC_MORE;
      }
    else
    if (HX_TEST_OPTION(opt_call_details))
      {
      options |= FNCCHK_CALLS_MORE;
      }
    else
    if (HX_TEST_OPTION(opt_memory))
      {
      options |= FNCCHK_MEMORY;
      }
    else
    if (HX_TEST_OPTION(opt_memory_more))
      {
      options |= FNCCHK_MEMORY_MORE;
      }
    else
    if (HX_TEST_OPTION(opt_pcalls))
      {
      options |= FNCCHK_CALLS_CALLED;
      }
    else
    if (HX_TEST_OPTION(opt_nm))
      {
      hxmessage("'%s': deprecated option. Ignored.", argv[i]);
      }
    else
    if (HX_TEST_OPTION(opt_addr2line))
      {
      hxmessage("'%s': deprecated option. Ignored.", argv[i]);
      }
    else
    if (HX_TEST_OPTION(opt_fullname))
      {
      options |= FNCCHK_FULLNAME;
      }
    else
    if (HX_TEST_OPTION(opt_sumprof))
      {
      if (!hx_cumulate_profile(argv[i+1], argv[i+2], argv[i+3]))
        {
        fprintf(stderr, "fncdump: error while adding profiles.\n");
        return(1);
        }
      return(0);
      i += 3;
      }
    else
    if (HX_TEST_OPTION(opt_sfile))
      {
      if ((statname = strdup(argv[i+1])) == NULL)
        {
	fprintf(stderr, "%s: Allocation error.\n", argv[0]);
	return(9);
	}
      i++;
      }
    else
    if (HX_TEST_OPTION(opt_only))
      {
      fonly = argv[i+1];
      i++;
      }
    else
    if (HX_TEST_OPTION(opt_not))
      {
      fnot = argv[i+1];
      i++;
      }
    else
    if (HX_TEST_OPTION(opt_sort))
      {
      sscanf(argv[i+1], "%d", &mode);
      if (mode < FNCCHK_SORT_MIN)
        mode = FNCCHK_SORT_MIN;
      if (mode > FNCCHK_SORT_MAX)
        mode = FNCCHK_SORT_MAX;
      i++;
      }
    else
    if (HX_TEST_OPTION(opt_usort))
      {
      sscanf(argv[i+1], "%d", &mode);
      if (mode < FNCCHK_SORT_MIN)
        mode = FNCCHK_SORT_MIN;
      if (mode > FNCCHK_SORT_MAX)
        mode = FNCCHK_SORT_MAX;
      mode = -mode;
      i++;
      }
    else
      {
      fprintf(stderr, "%s: Option '%s' ignored.\n", argv[0], argv[i]);
      }
    }

  if (filename == NULL)
    {
    fprintf(stderr, "No program to analyse!\n");
    return(4);
    }

  if (!fnccheck_dump(filename, statname, objlist, mode, options, fonly, fnot))
    {
    fprintf(stderr, "%s: Error during dump proccess.\n", argv[0]);
    return(3);
    }

  free(filename);
  if (statname != NULL)
    free(statname);
  return(0);
}
#undef HX_TEST_OPTION
