#include "fncgraph.h"




/* create a new node for a given function */
HX_Node *hx_create_node(HX_FDInfo *function)
{
  HX_Node *tmp;
  
  if (hx_malloc(tmp, 1) == NULL)
    {
    hxmemory();
    return(NULL);
    }
  if (hx_malloc(tmp->nexts, 1) == NULL)
    {
    hxmemory();
    free(tmp);
    return(NULL);
    }
  if (hx_malloc(tmp->prevs, 1) == NULL)
    {
    hxmemory();
    free(tmp->nexts);
    free(tmp);
    return(NULL);
    }
  tmp->nexts[0] = NULL;
  tmp->prevs[0] = NULL;
  tmp->function = function;
  tmp->keep     = 1;
  tmp->treated  = 0;
  tmp->tag      = 0;
  return(tmp);
}


/* add a child to a node */
int hx_add_child(HX_Node *node, HX_Node *child)
{
  int nb=0;

  while(node->nexts[nb] != NULL)
    nb++;
  if ((node->nexts = realloc(node->nexts, (nb+2)*sizeof(HX_Node*))) == NULL)
    {
    fprintf(stderr, "fncdump: Memory error.\n");
    return(0);
    }
  node->nexts[nb]   = child;
  node->nexts[nb+1] = NULL;
  
  return(1);
}

/* add a child to a node */
int hx_add_prev(HX_Node *node, HX_Node *prev)
{
  int nb=0;

  while(node->prevs[nb] != NULL)
    nb++;
  hx_reallocate_return(node->prevs, (nb+2), 0);
  node->prevs[nb]   = prev;
  node->prevs[nb+1] = NULL;
  
  return(1);
}



/* propagate given value (0/1) from the given node */ 
void hx_propagate_to_child(HX_Node*node, int val)
{
  int i;

  if (node->keep == val)
    return; /* still done */

  node->keep    = val;
  node->treated = 1;
  i = 0;
  while(node->nexts[i] != NULL)
    {
    hx_propagate_to_child(node->nexts[i], val);
    i++;
    }
}
/* retro-propagate given value (0/1) from the given node */ 
void hx_propagate_to_caller(HX_Node*node, int val)
{
  int i;

  if (node->keep == val)
    return; /* still done */

  node->keep = val;
  node->treated = 1;
  i = 0;
  while(node->prevs[i] != NULL)
    {
    hx_propagate_to_caller(node->prevs[i], val);
    i++;
    }
}

/* propagate given value (0/1) from the given node */ 
void hx_propagate_to_child_p(HX_Node*node)
{
  int i;

  if (node->treated == 1)
    return; /* still done */

  node->keep    = 1;
  node->treated = 1;
  i = 0;
  while(node->nexts[i] != NULL)
    {
    hx_propagate_to_child_p(node->nexts[i]);
    i++;
    }
}
/* retro-propagate given value (0/1) from the given node */ 
void hx_propagate_to_caller_p(HX_Node*node)
{
  int i;

  if (node->treated == 1)
    return; /* still done */

  node->keep    = 1;
  node->treated = 1;
  i = 0;
  while(node->prevs[i] != NULL)
    {
    hx_propagate_to_caller_p(node->prevs[i]);
    i++;
    }
}



/* propagate tags to nodes */
void hx_tag_nodes(HX_Node *node, int tag)
{
  int i;

  /* stop if node still tagged */
  if (node->tag != 0)
    return;
  /* tag the node */
  node->tag = tag;
  /* tag each child */
  i = 0;
  while(node->nexts[i] != NULL)
    {
    hx_tag_nodes(node->nexts[i], tag+1);
    i++;
    }
}


/* for debug */
void hx_display_recurs(int options, HX_Node *node, unsigned int nc)
{
  if (options & FNCCHK_NO_DECORATION)
    {
    printf("recursion %f \"%s\"\n", node->function->total, node->function->name);
    }
  else
    {
    printf("%d: Simple recursion:\n", nc);
    printf("  '%s' calls itself.\n", node->function->name);
    printf("  Total time for this function: %f\n", node->function->total);
    printf("\n");
    }
}
void hx_display_cycle(int options, HX_Node *end, HX_Node *start, unsigned int nc)
{
  if (options & FNCCHK_NO_DECORATION)
    {
    printf("cycle %f \"%s\" \"%s\"\n", start->function->total,
             start->function->name, end->function->name);
    }
  else
    {
    printf("%d: Cycle:\n", nc);
    printf("  cycle from '%s' to '%s'\n", start->function->name,
                                          end->function->name);
    printf("  Total time for this cycle: %f\n", start->function->total);
    printf("\n");
    }
}


/* compute list of cycles */
int hx_compute_cycles(int options, HX_Node **nodes, int nbinfo)
{
  int i, j;
  unsigned int ncycle=1;  /* current cycle */

  /* first, we "tag" the nodes in calling order */
  /* this is done for each entering point, that
     is to say each node without callers  */
  for(i=0; i<nbinfo; i++)
    {
    if (nodes[i]->prevs[0] == NULL)
      {
      hx_tag_nodes(nodes[i], 1);
      }
    }
  /* if parts of the program are pure cycles (i.e.
     a recursive function compiled alone in a
     profiled module) they are not treated with
     the previous method. So we re-do this job on
     non-marked nodes to be sure to not forget nodes */
  for(i=0; i<nbinfo; i++)
    {
    if (nodes[i]->tag == 0)
      {
      hx_tag_nodes(nodes[i], 1);
      }
    }

  /* a cycle appears if a node's child got a tag
     that is greater or equal to its tag */
  for(i=0; i<nbinfo; i++)
    {
    j = 0;
    while(nodes[i]->nexts[j] != NULL)
      {
      /* simple recursion */
      if (nodes[i]->tag == nodes[i]->nexts[j]->tag)
        {
	hx_display_recurs(options, nodes[i], ncycle);
	ncycle++;
	}
      /* cycle */
      else
      if (nodes[i]->tag > nodes[i]->nexts[j]->tag)
        {
	hx_display_cycle(options, nodes[i], nodes[i]->nexts[j], ncycle);
	ncycle++;
	}
      j++;
      }
    }

  return(1);
}

