#include "fncmem.h"

/* blocs of memory */
Avlnode *hx_mems=NULL;
unsigned int blocks;

/* internal: max stack size */
static int hx_max_mem_stack=0;

/* memory count. after counting, this is the memory leak */
long hx_nb_mem_used=0;
long hx_nb_mem_max=0;  /* gives the real MAX memory used */
/* length of the memory used */
long hx_nb_heap_size=0;

static void *hx_heap_start=(void*)-1;
static void *hx_heap_end=0;

static int
hxmem_index_cmp(HX_Mem *m1, HX_Mem *m2)
{
	return (char *)m1->ptr - (char *)m2->ptr;
}

static int
hxmem_index_ptr_cmp(void *ptr, HX_Mem *m)
{
	return (char *)ptr - (char *)m->ptr;
}

/* search for the given pointer in the list (starting at the end
    because a same address can be re-given by 'malloc' after
    beeing freed, so we must be sure that it is the LAST one).
   Need to search into 'reallocs' because the pointer can change */
static Avlnode *hx_search_pointer(void *ptr)
{
	return avl_find(hx_mems, ptr, (AVL_CMP) hxmem_index_ptr_cmp);
}

static void hx_update_heap(void *ptr, unsigned int size)
{
  hx_heap_start = HX_MIN(hx_heap_start, ptr);
  hx_heap_end   = HX_MAX(hx_heap_end,   ptr+size);
  hx_nb_heap_size = hx_heap_end - hx_heap_start;
}

/* add an invalid entry */
static void hx_add_error(Avlnode *av, void *where, void *ptr, void *newptr, unsigned int size,
                   HX_MemStack *stack, unsigned int nb)
{
  HX_Mem *me;

  hx_malloc(me, 1);
  memset(me, 0, sizeof(*me));
  me->ptr = ptr;

  me->block = blocks++;
  me->where_f = where;
  me->stack_f = stack;
  me->nb_stack_f = nb;

  me->func_a = newptr;
  me->size_a = size;
  if (!av)
    avl_insert( &hx_mems, me, (AVL_CMP)hxmem_index_cmp, avl_dup_error );
  else
    {
    me->anext = av->avl_data;
    av->avl_data = me;
    }
}

/* add new malloc entry */
static HX_Mem *hx_add_malloc(Avlnode *av, void *where, void *ptr, unsigned int size,
                   HX_MemStack *stack, unsigned int nb)
{
  HX_Mem *me;

  hx_malloc(me, 1);
  memset(me, 0, sizeof(*me));
  me->ptr = ptr;

  me->block = blocks++;
  me->where_a = where;
  me->size_a = size;
  me->stack_a = stack;
  me->nb_stack_a = nb;

  if (!av)
    av = hx_search_pointer(ptr);
  if (!av)
    avl_insert( &hx_mems, me, (AVL_CMP)hxmem_index_cmp, avl_dup_error );
  else
    {
    me->anext = av->avl_data;
    av->avl_data = me;
    }

  /* the current memory size */
  hx_nb_mem_used += size;
  /* store the MAX of memory allocated */
  hx_nb_mem_max = HX_MAX(hx_nb_mem_used, hx_nb_mem_max);
  /* compute the memory block size */
  hx_update_heap(ptr, size);
  return me;
}
static void hx_add_free(int options, void *where, void *ptr,
                   HX_MemStack *stack, unsigned int nb)
{
  HX_Mem *me;
  Avlnode *av;

  if (!(av = hx_search_pointer(ptr)))
    {
    hx_add_error(NULL, where, ptr, NULL, 0, stack, nb);
    return;
    }
  me = av->avl_data;
  /* this pointer was freed !!! */
  if (me->where_f)
    {
    hx_add_error(av, where, ptr, NULL, 0, stack, nb);
    return;
    }
  /* Dispose of call stack, we don't need it for non-leaks */
  free(stack);
  me->where_f = where;
  me->stack_f = NULL;
  me->nb_stack_f = 0;
  if (me->stack_a)
    {
    free(me->stack_a);
    me->stack_a = NULL;
    }
  me->nb_stack_a = 0;
  /* remove this block from total */
  hx_nb_mem_used -= me->size_a;
  /* Dispose of all previous call stacks. */
  for (me = me->rprev; me; me=me->rprev)
    {
    if (!me->where_a)
      continue;
    free(me->stack_a);
    me->stack_a = NULL;
    me->nb_stack_a = 0;
    }
  if (!(options & FNCCHK_MEMORY_MORE))
  {
    HX_Mem *m2, *pre = NULL;
    for (me = av->avl_data; me; me=m2)
    {
      m2 = me->rprev;
      if (!me->anext && !pre)
        avl_delete( &hx_mems, me->ptr, (AVL_CMP)hxmem_index_ptr_cmp );
      else if (pre)
        pre->anext = me->anext;
      else
        av->avl_data = me->anext;
      if (m2 && m2 != me->anext)
      {
        free(me);
        av = hx_search_pointer(m2->ptr);
	pre = NULL;
	for (me = av->avl_data; me; pre=me,me=me->anext)
	  if (me == m2)
	    break;
      } else
      {
        free(me);
      }
    }
  }
}
static void hx_add_realloc(void *where, void *ptr, void *newptr, unsigned int size,
                   HX_MemStack *stack, unsigned int nb)
{
  HX_Mem *me;
  Avlnode *av;

  if (!(av = hx_search_pointer(ptr)))
    {
    hx_add_error(NULL, where, ptr, newptr, size, stack, nb);
    return;
    }
  me = av->avl_data;
  /* this pointer was freed !!! */
  if (me->where_f)
    {
    hx_add_error(av, where, ptr, NULL, 0, stack, nb);
    return;
    }
  /* remove this block from total */
  hx_nb_mem_used -= me->size_a;

  me->rnext = hx_add_malloc(ptr == newptr ? av : NULL, where, newptr, size,
    stack, nb);
  me->rnext->rprev = me;
}

/* read the stack list from file */
static HX_MemStack *hx_read_call_stack(FILE *f, int *nb)
{
  HX_MemStack *tmp;
  char buf[1024], *p1, *p2;
  int i;

  /* nothing to read */
  if (hx_max_mem_stack <= 0)
    {
    *nb = 0;
    return(NULL);
    }

  /* read elements */
  if ((tmp = malloc((hx_max_mem_stack)*sizeof(HX_MemStack))) == NULL)
    {
    *nb = 0;
    hxmessage("allocation error while reading memory profile.");
    return(NULL);
    }
  *nb = 0;
  fgets(buf, sizeof(buf), f);
  p1 = buf;
  for(i=0; i<hx_max_mem_stack; i++)
    {
    tmp[i].addr = (void *)strtoul(p1, &p2, 16);
    p1 = p2;
    tmp[i].name = NULL;
    tmp[i].file = NULL;
    p1 = p2;
    if (tmp[i].addr == (void*)1)
      {
      tmp[i].addr = (void *)strtoul(p1, &p2, 16);
      p1 = p2;
      }
    if (tmp[i].addr == NULL)
      {
      break;  /* end of list */
      }
    (*nb)++;
    }
  return(tmp);
}

/* read a memory trace file */
int hx_read_memory_trace(int options, char *file)
{
  FILE *f;
  int nb, cont, i, j, ok;
  unsigned int size;
  void *where, *ptr, *newptr;
  char sop[64], buf[HX_MAX_BUFFER], buf2[HX_MAX_BUFFER], buf3[HX_MAX_BUFFER];
  HX_MemStack *cstack;
  unsigned int nbs;

  if ((f = fopen(file, "r")) == NULL)
     {
     hxmessage("can't open memory file '%s'.", file);
     hxmessage("memory operations ignored.");
     return(0);
     }
  /* read the file and generate memory datas */
  fgets(buf, 1023, f);
  if (buf[strlen(buf)-1] == '\n')
    buf[strlen(buf)-1] = '\0';
  sscanf(buf, "%1023s%1023s%d", buf2, buf3, &hx_max_mem_stack);
  if ((strcmp(buf2, "=") != 0)||(strcmp(buf3, "Start") != 0))
    {/* not a valid file! */
    hxmessage("bad memory file (%s).", file);
    hxmessage("memory operations ignored.");
    fclose(f);
    return(0);
    }
  cont = 1;
  nb = 0;
  do
    {
    /* first symbol */
    fgets(buf, HX_MAX_BUFFER, f);
    if (feof(f))
      {
      cont = 0;
      break;
      }
    nb++;
    /* no more used. just here for compatibility */
    /* now just tested by the end-of-file */
    if (buf[0] == '=')
      {/* MUST be the "= End" */
      cont = 0;
      }
    else
    if (buf[0] == '+')
      {
      sscanf(buf+2, "%x %x %0xu", &where, &ptr, &size);
      /* now read the corresponding call stack */
      cstack = hx_read_call_stack(f, &nbs);
      /* add the malloc in the list */
      hx_add_malloc(NULL, where, ptr, size, cstack, nbs);
      }
    else
    if (buf[0] == '-')
      {
      sscanf(buf+2, "%x %x", &where, &ptr);
      /* now read the corresponding call stack */
      cstack = hx_read_call_stack(f, &nbs);
      /* add the free in the list */
      hx_add_free(options, where, ptr, cstack, nbs);
      }
    else
    if (buf[0] == '>')
      {
      sscanf(buf+2, "%x %x %x %0xu", &where, &ptr, &newptr, &size);
      /* now read the corresponding call stack */
      cstack = hx_read_call_stack(f, &nbs);
      /* add the realloc in the list */
      hx_add_realloc(where, ptr, newptr, size, cstack, nbs);
      }
    else  /* ?!? */
      {
      hxmessage("bad start of line (%s) in file '%s'.", buf, file);
      continue;
      }
    /* end of file */
    if (cont && feof(f))
      {
      cont = 0;
      }
    }
  while(cont);
  fclose(f);

  /* done */
  return(1);
}
