#include "fncnames.h"


/* FIXME: this file is in the current directory, but is a part of
   binutils. But binutils don't installs it in standard include/
   directory! (needed for function cplus_demangle()) */
#include "demangle.h"


/* texts for special values (functions, files...) */
char *hx_ndef_function="<unknown_fnc>";
char *hx_ndef_file="<unknown_file>";
char *hx_ndef_object="<unknown_obj>";
char *hx_mem_function="<merror_fnc>";
char *hx_mem_file="<merror_file>";
char *hx_mem_object="<merror_obj>";
char *hx_before_function="<system_b>";
char *hx_after_function="<system_a>";
char *hx_baddr_function="<sys-:%p>";
char *hx_aaddr_function="<sys+:%p>";
char *hx_err_memory="<merror>";
char *hx_err_ndef="<error>";



/* struct to keep track of names for demangling purposes */
typedef struct
{
  char *name;
  int done;
}HXNym;

/* address of a line of a function */
typedef struct
{
  void *addr;
  unsigned int line;
}HXLin;

/* temp structure to extract names from .so files */
typedef struct
{
  void *addr;
  char *name;
  char *file;
  HXLin *lines;
  int nb_lines;
  flagword flag;  /* corresponding flag */
  HXNym *xname;
}HXSym;
HXSym *hx_syms=NULL;
int hx_nb_syms=0;


/* this structure is devoted to contains each address of
   symbols to be treated. corresponding name, file and line
   are specified here, in order to be updated by the program */
typedef struct
{
  HX_Loc *loc;
  HXNym *xname;	  /* address of master name ptr */
  int done;
}HXRequest;


/* add an entry to the request list */
HXRequest* hx_add_request(HXRequest *reqlist, int *nb, HX_Loc *loc)
{
  if (*nb == 0)
    {
    reqlist = malloc(sizeof(HXRequest));
    }
  else
    {
    reqlist = realloc(reqlist, sizeof(HXRequest)*((*nb)+1));
    }
  if (reqlist == NULL)
    {
    hxmessage("Recoverable memory error.");
    hxmessage("names extraction aborted.");
    return(NULL);
    }

  loc->name = NULL;
  loc->object = NULL;
  loc->file = NULL;
  loc->line = 0;
  loc->valid = 0;

  reqlist[*nb].loc = loc;
  reqlist[*nb].done = 0;
  reqlist[*nb].xname = NULL;

  (*nb)++;
  return(reqlist);
}



/* return pointer on beginning of file basename */
char *hx_get_basename(char *name)
{
  char *ptr;

  if (name == NULL)
    return(NULL);
  if (name[0] == 0)
    return(name);
  /* return the 1st '/' from the end */
  ptr = strrchr(name, '/');
  if (ptr && ptr[1])
    return ptr+1;
  return(name);
}


/* add a 'm' before the name */
int hx_compute_new_name(char *buf, char *oname)
{
  int i;

  /* search for the '/' (if exist) */
  for(i=strlen(oname)-1; i>=0; i--)
    {
    if (oname[i] == '/')
      break;
    }
  /* no '/'. just do m%s */
  if (oname[i] != '/')
    {
    sprintf(buf, "m%s", oname);
    return(1);
    }
  /* copy the starting elements */
  strncpy(buf, oname, i+1);
  buf[i+1] = '\0';
  strcat(buf, "m");
  strcat(buf, &(oname[i+1]));
  return(1);
}

void hx_demangle(HXNym *xname, int options)
{
  char *tmp;

  if (xname->done)
    return;
  xname->done = 1;
  if (options & FNCCHK_NO_DEMANGLE)
    return;
  tmp = cplus_demangle(xname->name, DMGL_ANSI | DMGL_PARAMS);
  if (tmp && tmp != xname->name)
    {
    free(xname->name);
    xname->name = tmp;
    }
}

int hx_open_bfd_file(char *name, bfd **core_bfd, int *core_num_syms,
                     asection **core_text_sect, asymbol ***core_syms,
		     int *core_min_insn_size, int *core_offset_to_code)
{
  *core_bfd = bfd_openr(name, 0);
  if (!(*core_bfd))
    {
    fprintf(stderr, "fncdump: Cant open %s\n", name);
    return(0);
    }
  /* check format */
  if (!bfd_check_format(*core_bfd, bfd_object))
    {
    fprintf(stderr, "fncdump: File %s is not an object.", name);
    return(0);
    }
  /* get TEXT section */
  *core_text_sect = bfd_get_section_by_name(*core_bfd, ".text");
  if (!(*core_text_sect))
    {
    *core_text_sect = bfd_get_section_by_name (*core_bfd, "$CODE$");
    if (!(*core_text_sect))
      {
      fprintf(stderr, "fncdump: No TEXT section in object %s.\n", name);
      return(0);
      }
    }
  /* read symbol table */
  *core_num_syms = bfd_get_symtab_upper_bound(*core_bfd);
  if (*core_num_syms < 0)
    {
    fprintf(stderr, "fncdump: %s\n", bfd_errmsg(bfd_get_error()));
    return(0);
    }
  *core_syms = (asymbol **) malloc(sizeof(asymbol*)*(*core_num_syms));
  if (*core_syms == NULL)
    {
    fprintf(stderr, "fncdump: Memory error while allocating %d bytes.\n", (int)sizeof(asymbol*)*(*core_num_syms));
    fprintf(stderr, "fncdump: Fatal error!\n");
    exit(9);
    }
  *core_num_syms = bfd_canonicalize_symtab(*core_bfd, *core_syms);
  if (*core_num_syms < 0)
    {
    free(*core_syms);
    fprintf(stderr, "fncdump: %s\n", bfd_errmsg(bfd_get_error()));
    return(0);
    }
  *core_min_insn_size = 1;
  *core_offset_to_code = 0;
  switch (bfd_get_arch(*core_bfd))
    {
    case bfd_arch_vax:
    case bfd_arch_tahoe:
      *core_offset_to_code = 2;
      break;
    case bfd_arch_alpha:
      *core_min_insn_size = 4;
      break;
    default:
      break;
    }
  return(1);
}

int hx_compare_names(const void *e1, const void *e2)
{
  char *n1 = ((HXSym *)e1)->name;
  char *n2 = ((HXSym *)e2)->name;
  if (!n1) n1 = "";
  if (!n2) n2 = "";
  return strcmp(n1, n2);
}
int hx_compare_pointers(const void *e1, const void *e2)
{
  return ((HXSym*)e1)->addr - ((HXSym*)e2)->addr;
}

/* sort in descending order */
int hx_compare_dynobj(const void *e1, const void *e2)
{
	HX_DynObj **d1 = (HX_DynObj **)e1, **d2 = (HX_DynObj **)e2;
  return (*d2)->base - (*d1)->base;
}

int hx_init_extract_dynamic(int core_num_syms, asymbol **core_syms, int sortn)
{
  int i, j;

  if ((hx_syms = malloc(sizeof(HXSym)*core_num_syms)) == NULL)
    {
    fprintf(stderr, "fncdump: Recoverable memory error.\n");
    return(0);
    }
  /* put addr/names in table */
  for(i=0,j=0; i<core_num_syms; i++)
    {
    if (core_syms[i]->flags & BSF_FILE) continue;
    hx_syms[j].name = (char*)bfd_asymbol_name(core_syms[i]);
    hx_syms[j].addr = (void*)bfd_asymbol_value(core_syms[i]);
    hx_syms[j].flag = core_syms[i]->flags;
    hx_syms[j].xname = NULL;
    hx_syms[j].lines = NULL;
    hx_syms[j].nb_lines = 0;
    hx_syms[j].file = NULL;
#if 0
printf("%p %s ", hx_syms[j].addr, hx_syms[j].name ? hx_syms[j].name : "??");
if (core_syms[i]->flags & BSF_GLOBAL)
  printf("BFD_GLOBAL ");
if (core_syms[i]->flags & BSF_LOCAL)
  printf("BSF_LOCAL ");
if (core_syms[i]->flags & BSF_FUNCTION)
  printf("BSF_FUNCTION ");
if (core_syms[i]->flags & BSF_WEAK)
  printf("BSF_WEAK ");
if (core_syms[i]->flags & BSF_INDIRECT)
  printf("BSF_INDIRECT ");
if (core_syms[i]->flags & BSF_FILE)
  printf("BSF_FILE ");
if (core_syms[i]->flags & BSF_DYNAMIC)
  printf("BSF_DYNAMIC ");
if (core_syms[i]->flags & BSF_OBJECT)
  printf("BSF_OBJECT ");
printf("\n");
#endif
    j++;
    }
  hx_nb_syms = j; /* core_num_syms; */
  /* sort it by pointer value */
  if (sortn)
    qsort(hx_syms, hx_nb_syms, sizeof(HXSym), hx_compare_pointers);
  else
    qsort(hx_syms, hx_nb_syms, sizeof(HXSym), hx_compare_names);
  return(1);
}
int hx_fini_extract_dynamic()
{
  if (hx_syms != NULL)
    free(hx_syms);
  hx_syms = NULL;
  hx_nb_syms = 0;
  return(1);
}
int hx_extract_dynamic(HXRequest *req, int *idx)
{
  int i, lo, hi;
  char *tmp, buffer[1024];

  req->loc->valid = 0;
  if (hx_syms == NULL)
    return(0);
  if (req->loc->addr < hx_syms[0].addr ||
      req->loc->addr > hx_syms[hx_nb_syms-1].addr)
    return(0);
  /* I may use a better search method :o) */
  /* Uses binsearch now! */
  for (lo=0, hi=hx_nb_syms-1; lo<=hi;)
    {
    i = (lo+hi) >> 1;
    if ((req->loc->addr >= hx_syms[i].addr) &&
        (req->loc->addr < hx_syms[i+1].addr))
      {
      if (!hx_syms[i].xname)
        {
	hx_syms[i].xname = (HXNym *)malloc(sizeof(HXNym));
	hx_syms[i].name = strdup(hx_syms[i].name);
	hx_syms[i].xname->name = hx_syms[i].name;
	hx_syms[i].xname->done = 0;
	}
      req->xname = hx_syms[i].xname;
      req->loc->name = hx_syms[i].name;
      req->loc->valid = 1;
      *idx = i;
      return(1);
      }
      if (hx_syms[i].addr > req->loc->addr)
        hi = i-1;
      else
        lo = i+1;
    }
  /* may not occur */
  return(0);
}

static char **files;
static int numfiles;

char *hx_dup_file(char *file, int options)
{
  int lo, hi, i, j;

  if (file == NULL)
    return hx_ndef_file;

  for (j=1, i=0, lo=0, hi=numfiles-1; lo<=hi;)
    {
    i = (lo+hi) >> 1;
    j = strcmp(files[i], file);
    if (j == 0)
      break;
    if (j>0)
      hi = i-1;
    else
      lo = i+1;
    }
  if (j)
    {
    files = realloc(files, (numfiles+1)*sizeof(char *));
    file = strdup(file);
    if (i>0 && j>0)
      --i;
    for (j=numfiles;j>i;j--)
      files[j] = files[j-1];
    files[j] = file;
    numfiles++;
    }
  file = files[i];
  if (!(options & FNCCHK_FULLNAME))
    return hx_get_basename(file);
  else
    return file;
}
char *hx_set_object(char *obj, int options)
{
  if (obj == NULL)
    return(hx_ndef_object);
  if (!(options & FNCCHK_FULLNAME))
    return hx_get_basename(obj);
  else
    return obj;
}

void hx_find_line(bfd *core_bfd, asection *core_text_sect,
	asymbol **core_syms, bfd_vma vaddr,HXRequest *req,int sym,int options)
{
  int i;

  for (i=0;i<hx_syms[sym].nb_lines;i++)
    {
    if (hx_syms[sym].lines[i].addr == req->loc->addr)
      break;
    }
  if (i == hx_syms[sym].nb_lines)
    {
    char *file, *func;

    hx_syms[sym].lines = realloc(hx_syms[sym].lines,
      (hx_syms[sym].nb_lines+1)*sizeof(HXLin));
    hx_syms[sym].nb_lines++;
    hx_syms[sym].lines[i].addr = req->loc->addr;
    bfd_find_nearest_line(core_bfd, core_text_sect, core_syms,
			  (bfd_vma)vaddr - core_text_sect->vma,
			  (const char**)&file, (const char**)&func,
			  &hx_syms[sym].lines[i].line);
    if (file && !hx_syms[sym].file)
      hx_syms[sym].file = hx_dup_file(file, options);
    }
  req->loc->file = hx_syms[sym].file;
  req->loc->line = hx_syms[sym].lines[i].line;
}

static int nb_reqlist;

int hx_add_memreq(HX_Mem *me, HXRequest **reqlist)
{
  int i;

  for(;me;me=me->anext)
  {
    if (me->where_a != NULL)
      {
      *reqlist = hx_add_request(*reqlist, &nb_reqlist,
                               (HX_Loc *)&me->where_a);
      }
    if (me->where_f != NULL)
      {
      *reqlist = hx_add_request(*reqlist, &nb_reqlist,
                               (HX_Loc *)&me->where_f);
      }
    /* names for memory stack */
    for(i=0; i<me->nb_stack_a; i++)
      {
      if (me->stack_a[i].addr == NULL)
	break;
      *reqlist = hx_add_request(*reqlist, &nb_reqlist,
                               &me->stack_a[i]);
      }
    for(i=0; i<me->nb_stack_f; i++)
      {
      if (me->stack_f[i].addr == NULL)
	break;
      *reqlist = hx_add_request(*reqlist, &nb_reqlist,
                               &me->stack_f[i]);
      }
  }
  return 0;
}

/* extract functions name from symbol addresses */
/* options: options of fncdump
   exec:    the executable name
   nbinfo:  number of entries in the symbol table
   objlist: NULL terminated list of dynamic objects used
            by the executable. the 1st element is the directory
	    corresponding to the executable
*/
int hx_extract_names(int options, char *exec, int nbinfo)
{
  int i, j, k;
  HXRequest *reqlist=NULL;
  char buffer[1024], tmpdir[1024];
  /* BFD data for exec */
  bfd *core_bfd;
  int core_min_insn_size;
  int core_offset_to_code;
  asection *core_text_sect;
  int core_num_syms;
  asymbol **core_syms;
  /* temp data */
  char *func_name;
  char *file_name;
  unsigned int line;
  int not_all_done=0, valid;
  char *object;

  /* first, we prepare the list of requested symbols */
  for(i=0; i<nbinfo; i++)
    {
    reqlist = hx_add_request(reqlist, &nb_reqlist, 
                             (HX_Loc *)&hx_dfinfo[i].function);
    }
  /* real call sites */
  for(i=0; i<nbinfo; i++)
    {
    for(j=0; j<hx_dfinfo[i].nb_of_callers; j++)
      {
      reqlist = hx_add_request(reqlist, &nb_reqlist,
                               &hx_dfinfo[i].rcb[j]);
      }
    }
  /* if present, read names for memory tracking */
  if (hx_mems)
  {
    avl_apply(hx_mems, (AVL_APPLY)hx_add_memreq, (void *)&reqlist, -1,
    	AVL_INORDER);
  }
  /* now open the executable */
  if (!hx_open_bfd_file(exec, &core_bfd, &core_num_syms,
                        &core_text_sect, &core_syms,
			&core_min_insn_size, &core_offset_to_code))
    {
    if (reqlist != NULL)
      free(reqlist);
    return(0);
    }
  /* we extract function names with our method, and after we extract
     the file:line information. */
  hx_init_extract_dynamic(core_num_syms, core_syms, 1);
  object = hx_set_object(exec, options);
  for(i=0; i<nb_reqlist; i++)
    {
    /* Did we get a valid address, and is it a definition? */
    if (hx_extract_dynamic(&reqlist[i], &j) &&
        (hx_syms[j].flag & (BSF_GLOBAL|BSF_LOCAL)))
      {
      reqlist[i].loc->object = object;
      reqlist[i].done = 1;
      hx_find_line(core_bfd,core_text_sect,core_syms,(bfd_vma)reqlist[i].loc->addr,
      		   &reqlist[i],j,options);
      }
    else
      not_all_done = 1;
    }
  hx_fini_extract_dynamic();
  /* close this bfd */
  bfd_close(core_bfd);
  free(core_syms);

  /* some symbols are not completed */
  /* the best way, here, would be to obtain the object name from
     where the symbol comes... as I dont know how to do that,
     I search in all objects */
  if (not_all_done && hx_nb_objects)
    {
    /* Sort the objects in descending order */
    qsort(hx_objects, hx_nb_objects, sizeof(HX_DynObj*), hx_compare_dynobj);

    for (k=0; not_all_done && k<hx_nb_objects; k++)
      {
      /* now open the object */
      if (!hx_open_bfd_file(hx_objects[k]->path, &core_bfd, &core_num_syms,
                            &core_text_sect, &core_syms,
			    &core_min_insn_size, &core_offset_to_code))
	{
	if (reqlist != NULL)
	  free(reqlist);
	return(0);
	}
      object = hx_set_object(hx_objects[k]->path, options);
      /* search for unmatched symbols */
      hx_init_extract_dynamic(core_num_syms, core_syms, 0);
      not_all_done = 0;
      for(i=0; i<nb_reqlist; i++)
        {
        if (reqlist[i].done == 0)
          {
	  /* Only check addresses that can reside in this module */
	  if (reqlist[i].loc->addr < hx_objects[k]->base) {
	    not_all_done = 1;
	    continue;
	  }
	  /* Search for a valid name but missing object/file info */
	  if (reqlist[i].loc->valid)
	    {
	    int lo, hi, n;
            /* search if a symbol matches the given element */
	    for(lo=0, hi=hx_nb_syms-1; lo<=hi;)
              {
	      j = (lo+hi)>>1;
	      n = strcmp(hx_syms[j].name, reqlist[i].loc->name);
	      if (n==0)
	        break;
	      if (n>0)
	        hi = j-1;
	      else
	        lo = j+1;
	      }
	    if (n==0 &&
/*	        (hx_syms[j].flag & (BSF_GLOBAL|BSF_WEAK))&& */
	  	(hx_syms[j].flag & BSF_FUNCTION))
              {/* it's the same name AND it is a function that is exported! */
	      reqlist[i].loc->object = object;
	      reqlist[i].done = 1;
	      hx_find_line(core_bfd, core_text_sect, core_syms,
	      		   (bfd_vma)reqlist[i].loc->addr, &reqlist[i], j, options);
	      }
	    else
	      not_all_done = 1;
	    }
	  else
	    not_all_done = 1;
          }
        }
      if (not_all_done)
        {
	/* Search for matching addresses, using the object base address */
        not_all_done = 0;
        qsort(hx_syms, hx_nb_syms, sizeof(HXSym), hx_compare_pointers);
        for (i=0; i<hx_nb_syms; i++)
          (long)hx_syms[i].addr += (long)hx_objects[k]->base;
        for (i=0; i<nb_reqlist; i++)
          {
	    if (reqlist[i].loc->valid)
	      continue;
	    /* Only check addresses that can reside in this module */
	    if (reqlist[i].loc->addr < hx_objects[k]->base)
	      {
	      not_all_done = 1;
	      continue;
	      }
	    if (hx_extract_dynamic(&reqlist[i], &j) &&
        	(hx_syms[j].flag & (BSF_GLOBAL|BSF_LOCAL|BSF_WEAK)))
	      {
	      reqlist[i].loc->object = object;
	      reqlist[i].done = 1;
	      hx_find_line(core_bfd, core_text_sect, core_syms,
			            (bfd_vma)(reqlist[i].loc->addr - hx_objects[k]->base),
				    &reqlist[i], j, options);
	      }
	    else
	      not_all_done = 1;
	  }
	}
      
      hx_fini_extract_dynamic();
      /* close this bfd */
      bfd_close(core_bfd);
      free(core_syms);
      }
    }

  /* add a default name for all unknown symbols */
  for(i=0; i<nb_reqlist; i++)
    {
    if (reqlist[i].done == 0)
      {
      reqlist[i].loc->file = hx_ndef_file;
      reqlist[i].loc->object = hx_ndef_object;
      }
    else
      {
      if (!reqlist[i].xname->done)
        hx_demangle(reqlist[i].xname, options);
      reqlist[i].loc->name = reqlist[i].xname->name;
      }
    }

  /* terminated */
  if (reqlist != NULL)
    free(reqlist);
  return(1);
}







