#ifndef __fnctools_h_
#define __fnctools_h_


/** this module manage allocations, errors and messages **/

/* WARNING: many macros in this file can only be used with GCC (_GNUC_)
      -> __FUNCTION__, args... & ##args */

/** includes **/
#include <stdarg.h>
#include "fnccommon.h"


/** return values **/
#define HX_OK         0
#define HX_ERR_OK     0
#define HX_ERR_MEM    9
#define HX_ERR_FILE   6
#define HX_ERR_STACK  5
#define HX_ERR_FUNC   4
#define HX_ERR_HASH   1
#define HX_ERR_NDEF   99


/** the dump program name (for messages) **/
extern char *hx_my_name;

/** texts to be displayed **/

/* names of sort types */
extern char *fncchk_sort_names[FNCCHK_SORT_MAX+2];
#define HX_SRT_NDEF "<?!?>"
#define HX_SRT_TOT  "Total time"
#define HX_SRT_LOC  "Local time"
#define HX_SRT_CALL "Number of calls"
#define HX_SRT_NAME "Function name"
#define HX_SRT_NON  "No sort"
/* title of sections */
#define HX_TTL_FLAT      "Flat profile:\n"
#define HX_TTL_CYCLES    "Cycles/recursions:\n"
#define HX_TTL_CALLG     "Call-graph:\n"
#define HX_TTL_MEM       "Memory blocks:\n"
#define HX_TTL_MEM2      "Memory leaks:\n"
#define HX_TTL_MINMAX    "Min/MAX time for functions:\n"
#define HX_TTL_DETAIL    "Details for functions:\n"



/** message functions **/
void hxmessage(char *format, ...);  /* display a msg */
void hxfatal(int ret);              /* fatal msg, with exit(ret) */
void hxfatal_msg();                 /* fatal msg */
#define hxmemory()  fprintf(stderr, "%s: Allocation error.", hx_my_name)



/** debug functions **/
void hx_real_printf(char*file, char*func, int line, int dlevel, char *format, ...);
#ifdef FNCCHK_DEBUG
#define hxprintf(lev,frm,args...)  hx_real_printf(__FILE__,__FUNCTION__,__LINE__,lev, frm, ##args)
#else
#define hxprintf(lev,frm,args...)  /* nothing */
#endif



/** allocation functions/macros **/
#define hx_malloc(var, nbel)   ((var) = malloc(sizeof(*(var))*(nbel)))
#define hx_realloc(var, nbel)   ((var) = realloc((var),sizeof(*(var))*(nbel)))
#define hx_smalloc(var, nbel)   (malloc(sizeof(*(var))*(nbel)))
#define hx_srealloc(var, nbel)   (realloc((var),sizeof(*(var))*(nbel)))
#define hx_allocate_fatal(var, nbel) \
  if (((var) = malloc((nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot allocate %d bytes at %s:%s:%d.\n", \
               hx_my_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    fprintf(stderr, "%s: Fatal error!\n", hx_my_name); \
    return(0); \
    }
#define hx_allocate_return(var, nbel, ret) \
  if (((var) = malloc((nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot allocate %d bytes at %s:%s:%d.\n", \
               hx_my_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    return(ret); \
    }
#define hx_reallocate_fatal(var, nbel) \
  if (((var) = realloc((var), (nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot reallocate %d bytes at %s:%s:%d.\n", \
               hx_my_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    fprintf(stderr, "%s: Fatal error!\n", hx_my_name); \
    return(0); \
    }
#define hx_reallocate_return(var, nbel, ret) \
  if (((var) = realloc((var), (nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot reallocate %d bytes at %s:%s:%d.\n", \
               hx_my_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    return(ret); \
    }
#define hx_reallocate(var, nbel) \
  if (((var) = realloc((var), (nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot reallocate %d bytes at %s:%s:%d.\n", \
               hx_my_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    }
#define hx_allocate(var, nbel) \
  if (((var) = malloc((nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot allocate %d bytes at %s:%s:%d.\n", \
               hx_my_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    }
#define hx_strdup_fatal(var, txt) \
  if (((var) = strdup(txt)) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot duplicate string at %s:%s:%d.\n", \
               hx_my_name, __FILE__,__FUNCTION__,__LINE__); \
    fprintf(stderr, "%s: Fatal error!\n", hx_my_name); \
    return(0); \
    }
#define hx_strdup(var, txt) \
  if (((var) = strdup(txt)) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot duplicate string at %s:%s:%d.\n", \
               hx_my_name, __FILE__,__FUNCTION__,__LINE__); \
    }




#endif /*__fnctools_h_*/
