/* Developped by Hexasoft (Y.Perret, yperret@ligim.univ-lyon1.fr) */

/*  Hexasoft, Mon December 4, 2000.
December 7,  2000: updating to V1.1.1
December 20, 2000: updating to V1.1.2
December 22, 2000: updating to V1.1.3
January 2,   2001: updating to V1.1.4
January 8,   2001: updating to V1.1.5
January 12,  2001: updating to V1.2
January 20,  2001: updating to V1.2.1
January 24,  2001: updating to V1.2.2
Feb 1,       2001: updating to V1.2.3
Feb 22,      2001: updating to V1.4
*/



/** to prevent compilation if not using GCC **/
#ifdef __GNUC__

#include <link.h>

/** includes **/
#include "fnccommon.h"   /* structures and so on */
#include "fnctime.h"     /* time manipulation */
#include "fncfunction.h" /* function table manipulation */
#include "fnctools.h"    /* various functions */
#include "fncmalloc.h"   /* memory tracing */

/** internationalization **/
#include <libintl.h>


/** function declarations **/
void __cyg_profile_func_exit(void *this_fn, void *call_site);



/** various variables used by the lib **/
/* these flags control profile activation during execution */
int hx_user_interact=1;
int hx_user_force_active=0;
/* the starting PID */
pid_t hx_starting_pid=(pid_t)0;
int hx_no_childs=0;  /* indicate to not dump childs */
/* time mode */
int hx_time_type=HX_TIME_EXT;
/* exec name: how to get it ? because I cant access argv[] here! */
/* perhaps using /proc/<PID>/ but is it portable ? */ 
char *hx_exec_name="<no_idea>";
/* the output file */
char *HX_DROP_FILE=FNCCHK_DEFAULT_FILE;
/* if true, indicate a REAL exit without treatments */
/* because as we register an atexit function, it
   will be called even for our internal exits, but
   these exits are emergency exits and may not be
   treated in the same way than a real exit */
int hx_internal_exit=0;
/* if true, indicate that exit is still done.
   this is used because C++ destructors are
   called AFTER the atexit call, and so after
   every data were freed.  */
int hx_exit_done=0;
/* info table */
HX_FInfo *hx_hx_dfinfo=NULL;
unsigned int hx_nb_finfos=0;
unsigned int hx_max_nb_finfos=0;  /* for realloc */
/* the last entry from ground */
unsigned int hx_last_entry=0;
/* stack of callers and entering times */
unsigned int hx_pos_in_stack=0;
unsigned int hx_pos_max_in_stack=0;  /* for realloc */
unsigned int hx_max_in_stack=0;      /* the MAX stack used */
struct timeval *hx_entering=NULL;
struct timeval hx_tnow, hx_tprev;
unsigned int *hx_who_is_before=NULL;
/* stack of real callers site (for file:line for memory stack) */
void **hx_mstack=NULL;
#ifdef HAVE_LIBPTHREAD
pthread_t *hx_pstack=NULL;
pthread_t hx_ptcur, hx_ptone;
#endif
int hx_dynamic_mode=0;  /* dynamique reallocations */
int hx_nb_of_realloc=0; /* count of reallocations */
/* hash table management */
HX_HashEntry *hx_hash_list=NULL;
unsigned int hx_nb_hash_list=0;


/* new: if true, indicates that a call to enter/exit happend
   AFTER atexit() call... (occur with static destructors)  */
int hx_maybe_in_destructor=0;
/* indicate if we accept to treat destructors */
int hx_accept_destructor=0;

/* dont display starting message */
int hx_dont_display_state=0;


/* allows thread support */
int hx_support_threads=0;

/* linker info */
static struct link_map *lm;

/** the function which dump the stat file **/
/* everything is freed after. it is so the last function */
static void hx_dump_to_file()
{
  FILE *f;
  unsigned int pos, j;
  void *fncfake=NULL;
  char buffer[HX_MAX_BUFFER];  /* to be changed */
#ifdef FNCCHK_DEBUG
  int i;
#endif

  /* if true, it is a call from a C++ destructor
     AFTER the prog exit. just forget it */
  if (hx_exit_done)
    return;

  /* is internal exit, no dump. it's an error exit */
  if (hx_internal_exit == 1)
    {
    if ((f = fopen(HX_DROP_FILE, "w")) != NULL)
      {/* make an empty stat file */
      fclose(f);
      }
    return; /* that's all */
    }
  else
  if (hx_internal_exit == 2)
    {/* hard exit! */
    return;
    }

  hxprintf(4, "()");

  /* debug level 2: display hash-table colisions */
#ifdef FNCCHK_DEBUG
  if (hx_debug_level & 2)
    {
    printf("# Hash table colisions:\n");
    for(i=0; i<hx_nb_hash_list; i++)
      {
      j = 0;
      while((j <HX_MAX_HASH_EL)&&(hx_hash_list[i][j].fnc != NULL))
        j++;
      printf("(%d:%d) ", i, j);
      }
    printf("\n# End\n");
    }
#endif

  /* if stack is not empty -> quit using 'exit()' */
  /* we call __..._leave() in order to make coherent
     informations available */
  /* dont apply this AFTER the standard exit */
  if (!hx_maybe_in_destructor)
    {
    if (hx_pos_in_stack > 0)
      {
      while(hx_pos_in_stack > 0)
	{
	pos = hx_who_is_before[hx_pos_in_stack-1];

	hxprintf(8, "force exit from function %u / %p", pos, hx_hx_dfinfo[pos].function);

	fncfake = hx_hx_dfinfo[pos].function;
	__cyg_profile_func_exit(fncfake, NULL);
	}
      }
    }

  /* if active, stop the memory tracking */
  if (hx_trace_memory)
    {
    hx_trace_suspend = 1;
    /* end of file */
    /* no more '= End'. now file is terminated on EOF */
    /*fprintf(hx_trace_file, "= End\n");*/
    if (!hx_accept_destructor)
      {
      fclose(hx_trace_file);
      }
    }

  /* change in the name generation */
  /* by default: if the current PID is not the starting PID
       a fork has been performed during execution, and we
       are (one of) the child(s). If FNCCHK_NOCHILD is set,
       nothing is done, else we add the PID to the name.  */
  if (hx_starting_pid == getpid())
    {/* no fork, or we are the father */
    sprintf(buffer, HX_DROP_FILE);
    }
  else
    {
    if (!hx_no_childs)
      {/* add the PID */
      sprintf(buffer, "%s_%d", HX_DROP_FILE, (int)getpid());
      }
    else
      {/* no dump, just quit */
        /* that was last call. freeing datas */
      if (hx_hx_dfinfo != NULL)
	free(hx_hx_dfinfo);
      if (hx_entering != NULL)
	free(hx_entering);
      if (hx_who_is_before != NULL)
	free(hx_who_is_before);
      hxprintf(4, "leaving ()");
      hxmessage("all done from child %u. No sfile generated.", (unsigned int)getpid());
      /* exit is done. never come again! */
      hx_exit_done = 1;
      return;
      }
    }

  if ((f = fopen(buffer, "w")) == NULL)
    {
    hxmessage("can't create stat-file '%s'.", buffer);
    return;
    }
  fprintf(f, "%s\n", FNCCHK_HEADER);
  fprintf(f, "%d\n", hx_nb_finfos);
  for(pos=0; pos<hx_nb_finfos; pos++)
    {
    /* output format:
	 HEADER
	 # of elements
	 f1_ptr  #calls sec usec locsec locusec minsec minusec maxsec maxusec
	 f2_ptr  #calls sec usec locsec locusec minsec minusec maxsec maxusec
	 ...
	 # of realloc
	 stack_size func_table_size
	 nb_of_callers_f1 c1_ptr rc1_ptr nb_c1 c2_ptr rc2_ptr nb_c2 ...
	 nb_of_callers_f2 c1_ptr rc1_ptr nb_c1 c2_ptr rc2_ptr nb_c2 ...
	 ...
	 time_mode
	 exec_name
    */
    fprintf(f, "%p %lu %lu %lu %lu %lu %lu %lu %lu %lu\n",
                        	 hx_hx_dfinfo[pos].function,
                        	 hx_hx_dfinfo[pos].nb_calls,
                        	 hx_hx_dfinfo[pos].nb_sec,
                        	 hx_hx_dfinfo[pos].nb_usec,
                        	 hx_hx_dfinfo[pos].nb_locsec,
                        	 hx_hx_dfinfo[pos].nb_locusec,
                        	 hx_hx_dfinfo[pos].min_sec,
                        	 hx_hx_dfinfo[pos].min_usec,
                        	 hx_hx_dfinfo[pos].max_sec,
                        	 hx_hx_dfinfo[pos].max_usec);
    }
  fprintf(f, "%d\n", hx_nb_of_realloc);
  fprintf(f, "%u %u\n", hx_max_in_stack, hx_nb_finfos);
  for(pos=0; pos<hx_nb_finfos; pos++)
    {
    fprintf(f, "%u ", hx_hx_dfinfo[pos].nb_of_callers);
    for(j=0; j<hx_hx_dfinfo[pos].nb_of_callers; j++)
      {
      fprintf(f, "%p %p %#x", hx_hx_dfinfo[pos].called_by[j],
        hx_hx_dfinfo[pos].rcalled_by[j], hx_hx_dfinfo[pos].nb_called_by[j]);
      if (j+1 < hx_hx_dfinfo[pos].nb_of_callers)
        {
        putc(' ', f);
	}
      }
    fprintf(f, "\n");
    }
  fprintf(f, "%d\n", hx_time_type);
  fprintf(f, "%d\n", hx_trace_memory);
  fprintf(f, "%s\n", hx_exec_name);
  for (; lm; lm=lm->l_next)
    if (lm->l_addr)
      fprintf(f, "%p %s\n", lm->l_addr, lm->l_name);
  fclose(f);


 
  /* that was last call. freeing datas */
  if (!hx_accept_destructor)
    {
    if (hx_hx_dfinfo != NULL)
      free(hx_hx_dfinfo);
    if (hx_entering != NULL)
      free(hx_entering);
    if (hx_who_is_before != NULL)
      free(hx_who_is_before);
    if (hx_mstack != NULL)
      free(hx_mstack);
#ifdef HAVE_LIBPTHREAD
    if (hx_pstack != NULL)
      free(hx_pstack);
#endif
    }
  /* exit is done. never come again! */
  if (!hx_accept_destructor)
    hx_exit_done = 1;

  hxprintf(4, "leaving ()");

  /* display this only one time */
  if (hx_maybe_in_destructor == 0)
    {
    if (hx_starting_pid != getpid())
      {
      if (!hx_dont_display_state)
        hxmessage("all done from child %u (sfile: %s).", (unsigned int)getpid(), buffer);
      }
    else
      {
      if (!hx_dont_display_state)
        hxmessage("all done (sfile: %s).", buffer);
      }
    }
  /* restart memory tracing, because if a destructor is called AFTER
     this, we will need the lasts memory actions */
  if ((hx_trace_memory)&&(!hx_exit_done))
    {
    hx_trace_suspend = 0;
    }
  /* indicate that normal exit is done */
  hx_maybe_in_destructor = 1;
}
/* used to redirect signals to atexit function */
static void hx_sig_dump_to_file(int dummy)
{
  hx_dump_to_file();
  hx_internal_exit = 2;
  exit(0);
}


/** init of the module system **/

/* add a 'm' before the name */
int hx_compute_new_name(char *buf, char *oname)
{
  int i;

  /* search for the '/' (if exist) */
  for(i=strlen(oname)-1; i>=0; i--)
    {
    if (oname[i] == '/')
      break;
    }
  /* no '/'. just do m%s */
  if (oname[i] != '/')
    {
    sprintf(buf, "m%s", oname);
    return(1);
    }
  /* copy the starting elements */
  strncpy(buf, oname, i+1);
  buf[i+1] = '\0';
  strcat(buf, "m");
  strcat(buf, &(oname[i+1]));
  return(1);
}

#ifdef HAVE_LIBPTHREAD
pthread_mutex_t hx_mutex;
#endif

static int hx_init_checker(void)
{
  char *def_stack, *def_fnc, *buf, *buf2;
  char rcf[HX_MAX_BUFFER];
  char mbuf[HX_MAX_BUFFER], mbuf2[HX_MAX_BUFFER];
  int def_istack=HX_MAX_STACK, def_ifnc=HX_MAX_FUNC;
  int i, j;
  FILE *f;
#ifdef ElfW
  ElfW(Dyn) *dyn = _DYNAMIC;
  struct r_debug *rdb;
#endif

  hxprintf(4, "()");

  /* read env for user-defined values */
  if ((def_stack = getenv("FNCCHK_QUIET")) != NULL)
    {
    if (strcmp(def_stack, "0") != 0)
      {
      hx_dont_display_state = 1;
      }
    }

#ifndef ElfW
  dlinfo(RTLD_SELF, RTLD_DI_LINKMAP, &lm);
  for(; lm && lm->l_prev; lm=lm->l_prev);
#else

  /* try to find out about dynamically loaded objects */
  for (; dyn->d_tag != DT_NULL; dyn++)
    {
    if (dyn->d_tag == DT_DEBUG)
      {
      	rdb = (struct r_debug *)dyn->d_un.d_ptr;
	lm = rdb->r_map;
	break;
      }
    }
#endif

  /* rc file name, searched in local dir to allow multiple files */
  /* rc info are readen BEFORE shell variables. so you can have a
     default config, and overwrite specific parts by using shell
     variables if you need, w/o having to change rc file.        */
  sprintf(rcf, FNCCHK_RC_FILE, "./");
  if ((f = fopen(rcf, "r")) != NULL)
    {/* read the file */
    while(!feof(f))
      {
      /* syntax is SHELL_VARIABLE VALUE */
      /* please ALWAYS define a VALUE */
      /* this loop just do 'setenv(SHELL_VAR, VALUE)' w/o
           any checking of validity (appart that variables
	   start by FNCCHK_). */
      fscanf(f, "%1023s%1023s", mbuf, mbuf2);
      if (strncmp(mbuf, "FNCCHK_", 7) != 0)
        {
	hxmessage("rc: invalid variable '%s' in rc file '%s'.",
	                    mbuf, rcf);
	}
      else
        {
	/* don't overwrite */
	if (!getenv(mbuf))
	  {
	  char *str=malloc(strlen(mbuf)+strlen(mbuf2)+2);
	  if (str)
	    {
	    sprintf(str,"%s=%s",mbuf,mbuf2);
	    putenv(str);
	    }
	  }
	}
      }
    fclose(f);
    }

  /* NOTE: for all shell variables, a value of "0" means that
       the variable is not set. This allows to 'unset' a variable
       specified in the .rc file.  */

  /* SUSPEND indicate that all is desactivated until user
    uses the fnccheck_start_profile() function */
  if (!hx_user_force_active)
    if ((def_stack = getenv("FNCCHK_SUSPEND")) != NULL)
      {
      if (strcmp(def_stack, "0") != 0)
        {
	hx_user_interact = 0;
	/* dont consider that init is done. */
	return(1);
	}
      }

  /* just said that we are here... */
  if (!hx_dont_display_state)
    hxmessage("starting (V%s by H.Chu).", FNCCHK_FULL_VERSION);

  /* read env for user-defined values */
  if ((def_stack = getenv("FNCCHK_STOP")) != NULL)
    {
    if (strcmp(def_stack, "0") != 0)
      {
      if (!hx_dont_display_state)
        hxmessage("stopped by FNCCHK_STOP.");
      hx_exit_done = 1;
      return(0);
      }
    }

  /* store the starting PID */
  hx_starting_pid = getpid();

  /* register exit function */
  if (atexit(hx_dump_to_file) != 0)
    {
    hxmessage("init: can't register exit function.");
    hxfatal(HX_ERR_NDEF);
    }
  /* register kill signals to make emergency dumps */
  /* which signals need to be catched ? */
  signal(SIGHUP,  hx_sig_dump_to_file);
  signal(SIGINT,  hx_sig_dump_to_file);
  signal(SIGTERM, hx_sig_dump_to_file);
  signal(SIGQUIT, hx_sig_dump_to_file);
  signal(SIGPIPE, hx_sig_dump_to_file);
  signal(SIGTERM, hx_sig_dump_to_file);

  /* read env for user-defined values */
  if ((def_stack = getenv("FNCCHK_STACKSIZE")) != NULL)
    {
    sscanf(def_stack, "%d", &def_istack);
    if (def_istack < 16)  /* invalid value */
      def_istack = HX_MAX_STACK;
    }
  if ((def_fnc = getenv("FNCCHK_NBFUNC")) != NULL)
    {
    sscanf(def_fnc, "%d", &def_ifnc);
    if (def_ifnc < 16)  /* invalid value */
      def_ifnc = HX_MAX_FUNC;
    }
  if ((def_fnc = getenv("FNCCHK_DEBUG")) != NULL)
    {
    sscanf(def_fnc, "%d", &hx_debug_level);
    if (hx_debug_level < 0)  /* invalid value */
      hx_debug_level = 0;
    }
  if ((buf = getenv("FNCCHK_OUTFILE")) != NULL)
    {
    hx_strdup_fatal(buf2, buf);
    HX_DROP_FILE = buf2;
    }
  if ((buf = getenv("FNCCHK_TIME")) != NULL)
    {
    hx_time_type = hx_set_time_type(buf);
    }
  if ((def_fnc = getenv("FNCCHK_DYNAMIC")) != NULL)
    {
    if (strcmp(def_fnc, "0") != 0)
      hx_dynamic_mode = 1;
    }
  if ((def_fnc = getenv("FNCCHK_NOCHILD")) != NULL)
    {
    if (strcmp(def_fnc, "0") != 0)
      hx_no_childs = 1;
    }
  if ((def_fnc = getenv("FNCCHK_DESTRUCTOR")) != NULL)
    {
    if (strcmp(def_fnc, "0") != 0)
      hx_accept_destructor = 1;
    }
#ifdef HAVE_LIBPTHREAD
  if ((def_fnc = getenv("FNCCHK_THREADS")) != NULL)
    {
    if (strcmp(def_fnc, "0") != 0)
      hx_support_threads = 1;
      pthread_mutex_init(&hx_mutex, NULL);
      hx_allocate_fatal(hx_pstack, def_istack);
      hx_ptone = pthread_self();
    }
#endif
  hx_pos_max_in_stack = def_istack;
  hx_max_nb_finfos = def_ifnc;

  /* data allocations */
  hx_allocate_fatal(hx_hx_dfinfo, def_ifnc);
  hx_allocate_fatal(hx_entering, def_istack);
  hx_allocate_fatal(hx_who_is_before, def_istack);
  hx_allocate_fatal(hx_mstack, def_istack);
  hx_nb_hash_list = (unsigned int)(def_ifnc*1.6667);
  if ((hx_nb_hash_list % 2) == 0)
    hx_nb_hash_list++; /* dont ask why... I prefer odd size :) */
  hx_allocate_fatal(hx_hash_list, hx_nb_hash_list);
  /* init of hashtable */
  for(i=0; i<hx_nb_hash_list; i++)
    {
    for(j=0; j<HX_MAX_HASH_EL; j++)
      {
      hx_hash_list[i][j].fnc = NULL;
      hx_hash_list[i][j].pos = 0;
      }
    }
  /* init time */
  hx_init_time();
  /* this is at the end to prevent fnccheck mallocs to be treated */
  if ((def_fnc = getenv("FNCCHK_MEMORY")) != NULL)
    {
    if (strcmp(def_fnc, "0") != 0)
      {
      /* the output name for mallocs profile */
      hx_compute_new_name(mbuf, HX_DROP_FILE);
      if ((hx_trace_file = fopen(mbuf, "w")) == NULL)
	{
	hxmessage("cant create memory file '%s'.");
	hxmessage("FNCCHK_MEMORY ignored.");
	}
      else
	{/* init the memory tracing */
	setbuf(hx_trace_file, hx_trace_buf);
	hx_trace_memory = 1;
	/* stack size to drop */
	if ((def_fnc = getenv("FNCCHK_MEMORY_STACK")) != NULL)
          {
	  sscanf(def_fnc, "%d", &hx_trace_stack);
	  if (hx_trace_stack < 0)
	    hx_trace_stack = 0;
	  if (hx_trace_stack > def_istack)
	    hx_trace_stack = def_istack;
	  }
	/* file header */
	fprintf(hx_trace_file, "= Start %d\n", hx_trace_stack);
	/* init our hooks for memory trace */
	hx_init_malloc();
	}
      }
    }
  /* display active options */
  if (!hx_dont_display_state)
    {
    hxmessage("   stack size is: %d", def_istack);
    hxmessage("   functions table size is: %d", def_ifnc);
    if (hx_dynamic_mode)
      hxmessage("   reallocations allowed");
    if (hx_time_type == HX_TIME_EXT)
      hxmessage("   clock mode is 'external time'");
    if (hx_time_type == HX_TIME_CPU)
      hxmessage("   clock mode is 'CPU time'");
    if (hx_no_childs)
      hxmessage("   no dump for children");
    if (hx_trace_memory)
      hxmessage("   memory tracking enabled");
    }

  hxprintf(4, "leaving ()");
  /* returns '0' in all cases, instead of if FNCCHK_SUSPEND is set */
  return(0);
}


/** functions that manage profiles **/

/* get the 'entering' time and add it in the stack */
inline void hx_get_entering_time()
{
  /* stack error! */
  if (hx_pos_in_stack >= hx_pos_max_in_stack)
    {
    if (hx_dynamic_mode)
      {
      if (hx_trace_memory)  /* dont trace this */
        hx_trace_suspend = 1;
      hx_reallocate_fatal(hx_entering, hx_pos_max_in_stack*2);
      hx_nb_of_realloc++;
      hx_reallocate_fatal(hx_who_is_before, hx_pos_max_in_stack*2);
      hx_nb_of_realloc++;
      hx_reallocate_fatal(hx_mstack, hx_pos_max_in_stack*2);
      hx_nb_of_realloc++;
#ifdef HAVE_LIBPTHREAD
      if (hx_pstack) {
        hx_reallocate_fatal(hx_pstack, hx_pos_max_in_stack*2);
        hx_nb_of_realloc++;
      }
#endif
      hx_pos_max_in_stack *= 2;
      
      if (hx_trace_memory) /* start again memory tracing */
        hx_trace_suspend = 0;
      }
    else
      {
      hxmessage("stack is full.");
      hxmessage("switch to dynamic allocation by setting FNCCHK_DYNAMIC=1!");
      hxfatal(HX_ERR_STACK);
      }
    }
  /* read the time */
  hx_tprev = hx_tnow;
  hx_gettimeofday(&hx_tnow);
  hx_entering[hx_pos_in_stack] = hx_tnow;
  hx_pos_in_stack++;
  /* check the max used size of stack */
  if (hx_pos_in_stack > hx_max_in_stack)
    hx_max_in_stack = hx_pos_in_stack;
}



/* read the 'leaving' time and compute the delay elapsed
   (sec/usec) from the corrsponding 'enter' */
inline void hx_get_leaving_time(int cur, long *sec, long *usec)
{
  /* stack pop */
  if (hx_pos_in_stack > 0)
    hx_pos_in_stack--;

  /* read time */
  hx_tprev = hx_tnow;
  hx_gettimeofday(&hx_tnow);
  /* compute delay */
  if (hx_tnow.tv_sec == hx_entering[cur].tv_sec)
    {
    *sec = 0;
    *usec = hx_tnow.tv_usec - hx_entering[cur].tv_usec;
    }
  else
    {
    *sec = hx_tnow.tv_sec - hx_entering[cur].tv_sec - 1;
    *usec = hx_tnow.tv_usec + 1000000-hx_entering[cur].tv_usec;
    }
  if (*usec >= 1000000)
    {
    *sec += 1;
    *usec -= 1000000;
    }
}

#ifdef HAVE_LIBPTHREAD
#ifdef ElfW	/* Linux */
static struct _pthread_cleanup_buffer hx_pop_buffer;
#else		/* Solaris */
static _cleanup_t hx_pop_buffer;
#endif

static void hx_pop_thread(void *dummy)
{
  unsigned int pos;
  pthread_t pt;
  int i;

  pt = pthread_self();
  hxprintf(4, "(%d)", pt);
  for (i=hx_pos_in_stack-1; i>0; i--)
    {
    if (hx_pstack[i] != pt) continue;
    pos = hx_who_is_before[i];
    __cyg_profile_func_exit(hx_hx_dfinfo[pos].function, (void *)1);
    }
}
#endif

/** functions which treat entering/leaving **/


static int first=1;

/* function called when entering a function */
void __cyg_profile_func_enter(void *this_fn, void *call_site)
{
  unsigned int pos;
  int cur;
  long sec, usec;
#ifdef HAVE_LIBPTHREAD
  int pcur;
  pthread_t pt;
#endif

  /* if true, it is a call from a C++ destructor
     AFTER the prog exit. just forget it */
  if (hx_exit_done)
    return;

  /* at first call, prepare datas for treatments */
  if (first)
    {
    first = hx_init_checker();
    /* if FNCCHK_STOP, exit_done is set */
    if (hx_exit_done)
      return;
    if (first)
      return;  /* not a real 'init' */
    }

  /* user control */
  if (!hx_user_interact)
    {
    return;
    }

  HX_LOCK;

  hxprintf(4, "(%p, %p)", this_fn, call_site);

  /* is this function still recorded ? */
  if ((pos = hx_search_function(this_fn)) == 0)
    {
    hxmessage("'entering' treatment aborted due to unregistered function.");
    goto leave;
    }
  pos--;

  /* store the caller in our memory-stack */
  hx_mstack[hx_pos_in_stack] = call_site-1;
#ifdef HAVE_LIBPTHREAD
  if (hx_pstack) {
    pt = pthread_self();
    hx_pstack[hx_pos_in_stack] = pt;
  }
#endif

  /* if posinstack==0 -> store the function for future exits */
  if (hx_pos_in_stack == 0)
    hx_last_entry = pos;

  hx_hx_dfinfo[pos].nb_in_stack++;

  hxprintf(8, "function %u, %u times in the stack.",
                    pos, hx_hx_dfinfo[pos].nb_in_stack);

  cur = hx_pos_in_stack-1;
#ifdef HAVE_LIBPTHREAD
  pcur = cur;
  if (hx_pstack)
    {
    if (cur >= 0 && pt != hx_pstack[cur])
	{
	for (;cur>=0;cur--)
	  if (pt == hx_pstack[cur]) break;
	if (cur < 0 && pt != hx_ptone)  /* new thread */
	  {
#ifdef ElfW
	  _pthread_cleanup_push(&hx_pop_buffer, hx_pop_thread, NULL);
#else
	  __pthread_cleanup_push(hx_pop_thread, NULL, _getfp(), &hx_pop_buffer);
#endif
	  }
	}
#if 0
    for (;pcur>=0;pcur--)
      if (hx_pstack[pcur] == hx_ptcur) break;
#endif
    hx_ptcur = pt;
    }
#endif
  if (cur >= 0)
    {/* we know the caller */
    hx_register_caller(pos, hx_who_is_before[cur], call_site-1);
    }

#if 0 /* def HAVE_LIBPTHREAD */
  cur = pcur;
#endif

  /* just insert the caller */
  hx_who_is_before[hx_pos_in_stack] = pos;
  /* get the entering time (increase stack position) */
  hx_get_entering_time();
  /* NEW: treatments to compute local time for parent function.
     entering a new function means that the previous one is not
     curently active. So we can compute the elapsed local time
     for it (diff of "last active time" and current time) which
     is added to local time.  */
  if (cur >= 0)  /* >1 and not >0 because stack is still ++ */
    {
    hx_compute_elapsed_time(  /* last active time for calling function */
    	 hx_tprev.tv_sec, hx_tprev.tv_usec,
	 /* current time */
	 hx_tnow.tv_sec, hx_tnow.tv_usec,
	 &sec, &usec);
    /* add this time to local time of calling function */
    hx_hx_dfinfo[hx_who_is_before[cur]].nb_locsec  += sec;
    hx_hx_dfinfo[hx_who_is_before[cur]].nb_locusec += usec;
    }
  hxprintf(4, "leaving (%p, %p)", this_fn, call_site);
leave:
  HX_UNLOCK;
}

/* function called when exiting a function */
void __cyg_profile_func_exit(void *this_fn, void *call_site)
{
  unsigned int pos;
  int cur;
  long sec, usec;
  unsigned long int nasec, nausec;
#ifdef HAVE_LIBPTHREAD
  int ptdif = 0;
#endif

  /* if true, it is a call from a C++ destructor
     AFTER the prog exit. just forget it */
  if (hx_exit_done)
    return;

  /* user control */
  if (!hx_user_interact)
    {
    return;
    }

  HX_LOCK;

  hxprintf(4, "(%p, %p)", this_fn, call_site);

  /* which function ? (posinstack-1 because stack points on next entry) */
  cur = hx_pos_in_stack-1;
#ifdef HAVE_LIBPTHREAD
  if (hx_pstack && call_site)
    {
    hx_ptcur = pthread_self();
    if (hx_ptcur != hx_pstack[cur])
      {
      ptdif = 1;
      for (;cur>=0;cur--)
        if (hx_pstack[cur] == hx_ptcur) break;
      }
    }
#endif
  if (cur >= 0)
    pos = hx_who_is_before[cur];
  else
    pos = hx_last_entry;  /* exiting the last activated function from ground */

  /* we compute the elapsed time */
  hx_get_leaving_time(cur, &sec, &usec);

#ifdef HAVE_LIBPTHREAD
  if (ptdif)
    {
    for (;cur<hx_pos_in_stack;cur++)
      {
      hx_mstack[cur] = hx_mstack[cur+1];
      hx_pstack[cur] = hx_pstack[cur+1];
      hx_entering[cur] = hx_entering[cur+1];
      hx_who_is_before[cur] = hx_who_is_before[cur+1];
      }
    }
#endif
   
  /* NEW: compute last part of local time for the current function */
  hx_compute_elapsed_time(
       hx_tprev.tv_sec, hx_tprev.tv_usec,
       hx_tnow.tv_sec, hx_tnow.tv_usec, &nasec, &nausec);
  hx_hx_dfinfo[pos].nb_locsec  += nasec;
  hx_hx_dfinfo[pos].nb_locusec += nausec;

  hxprintf(8, "function is %u. elapsed time is %ld sec & %ld usec",
                 pos, sec, usec);

  /* we register info in the structure */
  /* MIN/MAX management. It is done BEFORE removing subtime for
     recursive functions. So the MAX time will indicate the real
     MAX time spend in this function. */
  if (hx_hx_dfinfo[pos].nb_calls == 0)
    { /* first call: init min and max */
    hx_hx_dfinfo[pos].min_sec  = sec;
    hx_hx_dfinfo[pos].min_usec = usec;
    hx_hx_dfinfo[pos].max_sec  = sec;
    hx_hx_dfinfo[pos].max_usec = usec;
    }
  else /* test for new min / max */
    {
    if (HX_IS_GREATER(sec, usec, hx_hx_dfinfo[pos].max_sec, hx_hx_dfinfo[pos].max_usec))
      {
      hx_hx_dfinfo[pos].max_sec  = sec;
      hx_hx_dfinfo[pos].max_usec = usec;
      }
    else /* if it is the MAX, it may not be the MIN ! */
    if (HX_IS_LOWER(sec, usec, hx_hx_dfinfo[pos].min_sec, hx_hx_dfinfo[pos].min_usec))
      {
      hx_hx_dfinfo[pos].min_sec  = sec;
      hx_hx_dfinfo[pos].min_usec = usec;
      }
    }
  /* one more call */
  hx_hx_dfinfo[pos].nb_calls++;
  /* remove an occurence in the stack */
  hx_hx_dfinfo[pos].nb_in_stack--;

  /* if it is not a recursive function, we perform standard treatments */
  if (hx_hx_dfinfo[pos].nb_in_stack == 0)
    {
    /* add the (total) time to the function */
    hx_hx_dfinfo[pos].nb_sec  += sec;
    hx_hx_dfinfo[pos].nb_usec += usec;
    }
  else /* several occurence of the function in the stack */
    {
    /* nothing to do, I think... beeing checked */
    }

  /* new: if true, indicates that we are treating destructors...
     we need to call again 'dump' function */
  if (hx_maybe_in_destructor)
    {
    hx_dump_to_file();
    }

  hxprintf(4, "leaving (%p, %p)", this_fn, call_site);
  HX_UNLOCK;
}


/** NEW: these functions allow you to interact with fnccheck
    during execution, by stopping ar re-starting treatments   **/

/* stop profiling. All treatments are stopped */
void fnccheck_stop_profile()
{
  hx_user_interact = 0;
}

/* re-start profiling. Standard behavior is re-activated. */
void fnccheck_start_profile()
{
  hx_user_interact     = 1;
  hx_user_force_active = 1;  /* ignore SUSPEND if activate */
}

/* force dumping of profile informations.
Note: you will be be able to use profile again during this
  execution (even with 'fnccheck_stop_profile').   */
void fnccheck_force_dump()
{
  hx_dump_to_file();
}





#endif  /* __GNUC__ */
