#ifndef __fnccommon_h_
#define __fnccommon_h_


/** includes **/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/times.h>
#include <math.h>
#include <unistd.h>
#include <signal.h>
#include "../include/fnccheck.h"


/* various definitions shared by all .c */



/** time management section **/
/* the kind of time used */
#define HX_TIME_EXT 0   /* absolute time */
#define HX_TIME_CPU 1   /* internal time (CPU) */
#define HX_TIME_SYS 2   /* CPU+SYSTEM times */



/** debug stuff **/
extern int hx_debug_level;
/* translate this to output to stderr */
#define hx_printf    if(hx_debug_level)    printf
#define hx_printf1   if(hx_debug_level&1)  printf
#define hx_printf2   if(hx_debug_level&2)  printf
#define hx_printf4   if(hx_debug_level&4)  printf
#define hx_printf8   if(hx_debug_level&8)  printf
#define hx_printf16  if(hx_debug_level&16) printf
#define hx_printf32  if(hx_debug_level&32) printf



/** usefull MACROS and values **/
/* I think a macro to do this exists in time.h... */
#define HX_IS_GREATER(sec,usec,lsec,lusec)  (sec>lsec?1:(sec<lsec?0:(usec>lusec?1:0)))
#define HX_IS_LOWER(sec,usec,lsec,lusec)    (sec<lsec?1:(sec>lsec?0:(usec<lusec?1:0)))
/* string buffers */
#define HX_MAX_BUFFER 1024
/* 1 usec */
#define EPS 0.000001



/** function stat structure and associate **/
/* not really nice. The max number of callers for each
   function that are registered. */
#define HX_MAX_CALLER 64
/* structure of a function registration */
typedef struct
{
  void *function;  /* the pointer to the function */
  unsigned long int nb_calls; /* total # of calls */
  unsigned long int nb_sec;   /* total # of secs */
  unsigned long int nb_usec;  /* total # of usecs */
  unsigned int recurs_count;  /* recursive counts */
  /* local time of the function */
  unsigned long int nb_locsec;   /* # of secs */
  unsigned long int nb_locusec;  /* # of usecs */
  /* min and max time for the function */
  unsigned long int min_sec;   /* # of secs */
  unsigned long int min_usec;       /* # of usecs */
  unsigned long int max_sec;   /* # of secs */
  unsigned long int max_usec;       /* # of usecs */
  void * called_by[HX_MAX_CALLER];  /* list of callers */
  void * rcalled_by[HX_MAX_CALLER]; /* list of real callers (for file/line) */
  unsigned int nb_called_by[HX_MAX_CALLER];	/* # calls, per caller */
  unsigned int nb_of_callers;  /* pos in previous tab */
  unsigned int nb_in_stack;    /* # of insert in the stack */
}HX_FInfo;
/* info table */
#define HX_MAX_FUNC 1024
extern HX_FInfo *hx_hx_dfinfo;
extern unsigned int hx_nb_finfos;
extern unsigned int hx_max_nb_finfos;  /* for realloc */



/** call stack management **/
/* stack of callers and entering times */
#define HX_MAX_STACK 1000000
extern unsigned int hx_pos_in_stack;
extern unsigned int hx_pos_max_in_stack;  /* for realloc */
extern unsigned int hx_max_in_stack;      /* the MAX stack used */
extern struct timeval *hx_entering;
extern unsigned int *hx_who_is_before;
extern void **hx_mstack;



/** hash-table **/
typedef struct
{
  void *fnc;        /* function stored */
  unsigned int pos; /* pos int the list */
}HX_HashEl;
/* got a better way to prevent mallocs ? */
#define HX_MAX_HASH_EL 32  /* max collisions */
typedef HX_HashEl HX_HashEntry[HX_MAX_HASH_EL];
extern HX_HashEntry *hx_hash_list;
extern unsigned int hx_nb_hash_list;



/** various execution information and options **/
/* these flags control profile activation during execution */
extern int hx_user_interact;
extern int hx_user_force_active;
/* the starting PID */
extern pid_t hx_starting_pid;
extern int hx_no_childs;  /* indicate to not dump childs */
/* do we use memory tracking ? */
extern int hx_trace_memory;
/* exec name */
extern char *hx_exec_name;
/* the output file */
extern char *HX_DROP_FILE;
extern int hx_dynamic_mode;  /* dynamique reallocations */
extern int hx_nb_of_realloc; /* count of reallocations */
/* special flags to prevent multiple exits */
extern int hx_internal_exit;
extern int hx_exit_done;
/* time mode */
extern int hx_time_type;

/* locking for thread support? */
extern int hx_support_threads;
#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
extern pthread_mutex_t hx_mutex;
extern pthread_t *hx_pstack;
#define	HX_LOCK	if (hx_support_threads) pthread_mutex_lock(&hx_mutex)
#define	HX_UNLOCK	if (hx_support_threads) pthread_mutex_unlock(&hx_mutex)
#else
#define	HX_LOCK
#define	HX_UNLOCK
#endif

#endif /*__fnccommon_h_*/
