#ifndef __fnctools_h_
#define __fnctools_h_


/** this module manage allocations, errors and messages **/

/* WARNING: many macros in this file can only be used with GCC (_GNUC_)
      -> __FUNCTION__, args... & ##args */

/** includes **/
#include <stdarg.h>
#include "fnccommon.h"


/** return values **/
#define HX_OK         0
#define HX_ERR_OK     0
#define HX_ERR_MEM    9
#define HX_ERR_FILE   6
#define HX_ERR_STACK  5
#define HX_ERR_FUNC   4
#define HX_ERR_HASH   1
#define HX_ERR_NDEF   99


/** the library name for messages **/
extern char *hx_lib_name;

/** texts to be displayed **/



/** message functions **/
void hxmessage(char *format, ...);  /* display a msg */
void hxfatal(int ret);              /* fatal msg, with exit(ret) */

/** allocation functions/macros **/
#define hx_allocate_fatal(var, nbel) \
  if (((var) = malloc((nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot allocate %d bytes at %s:%s:%d.\n", \
               hx_lib_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    hxfatal(HX_ERR_MEM); \
    }
#define hx_reallocate_fatal(var, nbel) \
  if (((var) = realloc((var), (nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot reallocate %d bytes at %s:%s:%d.\n", \
               hx_lib_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    hxfatal(HX_ERR_MEM); \
    }
#define hx_allocate(var, nbel) \
  if (((var) = malloc((nbel)*(sizeof(*(var))))) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot allocate %d bytes at %s:%s:%d.\n", \
               hx_lib_name, ((nbel)*(sizeof(*(var)))), __FILE__,__FUNCTION__,__LINE__); \
    }
#define hx_strdup_fatal(var, txt) \
  if (((var) = strdup(txt)) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot duplicate string at %s:%s:%d.\n", \
               hx_lib_name, __FILE__,__FUNCTION__,__LINE__); \
    hxfatal(HX_ERR_MEM); \
    }
#define hx_strdup(var, txt) \
  if (((var) = strdup(txt)) == NULL) \
    { \
    fprintf(stderr, "%s: Cannot duplicate string at %s:%s:%d.\n", \
               hx_lib_name, __FILE__,__FUNCTION__,__LINE__); \
    }



/** debug functions **/
void hx_real_printf(char*file, char*func, int line, int dlevel, char *format, ...);
#ifdef FNCCHK_DEBUG
#define hxprintf(lev,frm,args...)  hx_real_printf(__FILE__,__FUNCTION__,__LINE__,lev, frm, ##args)
#else
#define hxprintf(lev,frm,args...)  /* nothing */
#endif



#endif /*__fnctools_h_*/
